import logging
import pathlib
import platform
import sys
import argparse
import json
import jsmin
from lussac.core import LussacData, LussacPipeline, LussacSpikeSorter


def parse_arguments(args: list | None) -> str:
	"""
	Parses the arguments given when launching Lussac.

	@param args: list
		All the arguments to parse.
	@return params_file: str
		Path to the file containing Lussac's parameters.
	"""

	parser = argparse.ArgumentParser()
	parser.add_argument("params_file", help="Path to the json file containing Lussac's parameters.")
	args = parser.parse_args(args)

	return args.params_file


def load_json(filename: str) -> dict:
	"""
	Loads the JSON parameters file and returns its content.

	@param filename: str
		Path to the file containing Lussac's parameters.
	@return params: dict
		Lussac's parameters.
	"""

	folder = pathlib.Path(filename).parent
	with open(filename) as json_file:
		minified = jsmin.jsmin(json_file.read())  # Parses out comments.
		minified = minified.replace("$PARAMS_FOLDER", str(folder.absolute()))
		if platform.system() == "Windows":  # pragma: no cover (OS specific).
			minified = minified.replace("\\", "\\\\")

		return json.loads(minified)


def main() -> None:  # pragma: no cover
	"""
	The main function to execute Lussac.
	"""

	# STEP 0: Loading the parameters into the LussacData object.
	params_file = parse_arguments(sys.argv[1:])
	params = load_json(params_file)
	data = LussacData.create_from_params(params)

	# STEP 1: Running the spike sorting.
	if 'spike_sorting' in params:
		logging.info("Running spike-sorting algorithms:\n")
		for name, params0 in params['spike_sorting'].items():
			spike_sorter = LussacSpikeSorter(data.recording, name)
			data.sortings[name] = spike_sorter.launch(params0)

	# STEP 2: Running the pipeline.
	pipeline = LussacPipeline(data)
	pipeline.launch()


if __name__ == "__main__":  # pragma: no cover
	main()



from typing import Any
from overrides import override
import numpy as np
import plotly.graph_objects as go
import scipy.signal
from lussac.core import MonoSortingModule
import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.curation as scur
import spikeinterface.postprocessing as spost


class AlignUnits(MonoSortingModule):
	"""
	Align the units so that their peak is at t=0 in the spike train.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'wvf_extraction': {
				'ms_before': 2.0,
				'ms_after': 2.0,
				'max_spikes_per_unit': 1_000
			},
			'filter': [200, 5000],
			'threshold': 0.5,
			'check_next': 10
		}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		templates, wvf_extractor, margin = self.get_templates(params['wvf_extraction'], params['filter'], return_extractor=True)
		best_channels = np.argmax(np.max(np.abs(templates), axis=1), axis=1)
		templates = templates[np.arange(templates.shape[0]), :, best_channels]  # Only take the best channel for each unit.

		shifts = self.get_units_shift(templates, wvf_extractor.nbefore - margin, params['threshold'], params['check_next'])
		self._plot_alignment(templates, wvf_extractor.nbefore - margin, shifts, params['threshold'])

		return self.shift_sorting(self.recording, self.sorting, {self.sorting.unit_ids[i]: -shifts[i] for i in range(len(shifts))})

	@staticmethod
	def shift_sorting(recording: si.BaseRecording, sorting: si.BaseSorting, shift: dict[Any, int]) -> si.BaseSorting:
		"""
		Shifts the spike train of the sorting by shift time samples.

		@param recording: si.BaseRecording
			The recording object (needed to check for spike going over the edges).
		@param sorting: si.BaseSorting
			The sorting to shift.
		@param shift: dict[Any, int]
			A dict mapping the unit id to its shift.
		@return aligned_sorting: si.BaseSorting
			The shifted sorting.
		"""

		aligned_sorting = spost.align_sorting(sorting, shift)
		aligned_sorting = scur.remove_excess_spikes(aligned_sorting, recording)

		return aligned_sorting

	@staticmethod
	def get_units_shift(templates: np.ndarray, nbefore: int, threshold: float, check_next: int) -> np.ndarray:
		"""
		Computes the shift between nbefore and the peak of the template.

		@param templates: np.ndarray (n_units, n_samples)
			The template for each unit.
		@param nbefore: int
			The index of t=0 in the templates.
		@param threshold: float
			First the first peak that is at least threshold*max.
		@param check_next: int
			Checks the next x time samples after the first peak, and take the highest one.
		@return units_shift: np.ndarray[int16] (n_units)
			The shift
		"""

		centers = np.empty(templates.shape[0], dtype=np.int16)

		for i, template in enumerate(templates):
			template = np.abs(template)
			peaks, _ = scipy.signal.find_peaks(template, height=threshold * np.max(template))

			if len(peaks) == 0:  # Can happen if the maximum is at the very start of end of the template.
				peaks = [np.argmax(template)]

			centers[i] = peaks[0] + np.argmax(template[peaks[0]:peaks[0] + check_next + 1])

		return centers - nbefore

	def _plot_alignment(self, templates: np.ndarray, nbefore: int, shifts: np.ndarray, threshold_ratio: float) -> None:
		"""
		Plots the template for each unit with the new center.

		@param templates: np.ndarray (n_units, n_samples)
			The template for each unit on its best channel.
		@param nbefore: int
			The index of t=0 in the templates.
		@param shifts: np.ndarray (n_units)
			The new center for each unit.
		@param threshold_ratio: float
			The threshold parameter used to compute the shifts.
		"""

		xaxis = (np.arange(templates.shape[1]) - nbefore) / utils.Utils.sampling_frequency * 1e3
		fig = go.Figure()
		labels = []
		args = []

		old_center_line = {'x0': 0, 'x1': 0, 'y0': 0, 'y1': 1, 'xref': 'x', 'yref': 'y domain', 'line': {'dash': "dash", 'color': "GoldenRod"}}  #, 'annotation': {'text': "Old Center", 'textangle': -90}}
		fig.add_shape(**old_center_line)
		fig.update_xaxes(title_text="Time relative to old center (ms)")
		
		for i in range(templates.shape[0]):
			unit_id = self.sorting.unit_ids[i]
			shift = shifts[i] / utils.Utils.sampling_frequency * 1e3
			threshold = threshold_ratio * np.max(np.abs(templates[i]))

			fig.add_trace(go.Scatter(
				x=xaxis,
				y=templates[i],
				mode="lines",
				marker_color="CornflowerBlue"
			))

			new_center_line = {'x0': shift, 'x1': shift, 'y0': 0, 'y1': 1, 'xref': 'x', 'yref': 'y domain', 'line': {'dash': "dash", 'color': "Crimson"}}  #, 'annotation': {'text': "New Center", 'textangle': -90}}
			threshold_line_up = {'x0': 0, 'x1': 1, 'y0': threshold, 'y1': threshold, 'xref': 'x domain', 'yref': 'y', 'line': {'dash': "dash", 'color': "LightSeaGreen"}, 'opacity': 0.4}
			threshold_line_down = {'x0': 0, 'x1': 1, 'y0': -threshold, 'y1': -threshold, 'xref': 'x domain', 'yref': 'y', 'line': {'dash': "dash", 'color': "LightSeaGreen"}, 'opacity': 0.4}
			if i == 0:
				fig.add_shape(**new_center_line)
				fig.add_shape(**threshold_line_up)
				fig.add_shape(**threshold_line_down)

			labels.append(f"Unit {unit_id}")
			args.append({
				'title_text': f"Unit {unit_id}",
				'shapes': [old_center_line, new_center_line, threshold_line_up, threshold_line_down]
			})

		utils.plot_sliders(fig, 1, labels, f"{self.logs_folder}/alignment", args)



import os
import importlib
import inspect
from typing import Any, Type
from lussac.core import LussacModule


class ModuleFactory:
	"""
	This is a class to load all the modules in the "modules" folder.
	"""

	module_classes: dict[str, Type[LussacModule]]

	def __init__(self) -> None:
		"""
		Creates a new ModuleFactory instance.
		Loads all the modules from the "modules" folder.
		"""

		self.module_classes = self._load_modules()

	@staticmethod
	def _load_modules() -> dict[str, Type[LussacModule]]:
		"""
		Loads all the modules from the "modules" folder.

		@return modules: dict[str, Type[LussacModule]]
			All the modules classes.
		"""

		modules = {}

		for module_file in os.listdir(os.path.join(os.path.dirname(__file__), '../modules')):
			if not module_file.endswith('.py') or module_file.startswith('_'):
				continue

			module_name = module_file[:-3]
			modules[module_name] = ModuleFactory._get_module_member(f"lussac.modules.{module_name}")

		return modules

	@staticmethod
	def _get_module_member(module_name: str) -> Type[LussacModule]:
		"""
		Returns the module class from its path.

		@param module_name: str
			Path to the module file as would be used in an import statement.
		@return: module_class: Type[LussacModule]
			The module class (not instantiated!).
		"""

		module = importlib.import_module(module_name)
		members = inspect.getmembers(module, ModuleFactory._is_member_lussac_module)
		members = [member for member in members if member[1].__module__ == module_name]

		if len(members) == 0:
			raise Exception(f"Error: Couldn't find a module class for module '{module_name}'.")
		if len(members) > 1:
			raise Exception(f"Error: Found multiple module classes for module '{module_name}'.\n{members}")

		return members[0][1]

	@staticmethod
	def _is_member_lussac_module(member: Any) -> bool:
		"""
		Checks if a member (from inspect.getmembers) is a LussacModule class.

		@param member: Any
			The member to check.
		@return is_lussac_module: bool
			True if the member is a LussacModule class.
		"""

		return inspect.isclass(member) and issubclass(member, LussacModule) and not inspect.isabstract(member)

	def get_module(self, name: str) -> Type[LussacModule]:
		"""
		Gets a module class by its name.

		@param name: str
			The name of the module.
		@return module: Type[LussacModule]
			The module class.
		"""

		if name not in self.module_classes:
			raise ValueError(f"Error: Module '{name}' not found.\nThe loaded modules are: {list(self.module_classes.keys())}")

		return self.module_classes[name]



from typing import Any
from overrides import override
from lussac.core import MonoSortingModule
import spikeinterface.core as si
import spikeinterface.postprocessing as spost


class ExportToSIGUI(MonoSortingModule):
	"""
	Exports the sorting data as a waveform extractor (for SpikeInterface GUI).
	"""

	export_sortings = False

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'wvf_extraction': {
				'ms_before': 1.0,
				'ms_after': 3.0,
				'max_spikes_per_unit': 1000,
				'return_scaled': True,
				'allow_unfiltered': True
			},
			'spike_amplitudes': {

			},
			'principal_components': False
		}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		path = self._format_output_path(params['path'])
		wvf_extractor = si.extract_waveforms(self.recording, self.sorting, path, **params['wvf_extraction'])

		if isinstance(params['spike_amplitudes'], dict):
			spost.compute_spike_amplitudes(wvf_extractor, **params['spike_amplitudes'])
		if isinstance(params['principal_components'], dict):
			spost.compute_principal_components(wvf_extractor, **params['principal_components'])

		return self.sorting

	def _format_output_path(self, path: str) -> str:
		"""
		Formats the path to the output folder.
		If there is only 1 sorting, it remains unchanged.
		However, if there are multiple sorting, the sorting's name is added.

		@param path: str
			The initial path to the output folder.
		@return output_path: str
			The formatted output path.
		"""

		if self.data.data.num_sortings > 1:
			path = f"{path}/{self.data.name}"

		return path



import inspect
import math
from typing import Any, Callable, TypeVar
import scipy.interpolate
import scipy.stats
import numba
import numpy as np
import numpy.typing as npt
from .variables import Utils
from spikeinterface.curation.auto_merge import get_unit_adaptive_window, normalize_correlogram
from spikeinterface.postprocessing.correlograms import _compute_crosscorr_numba


T = TypeVar("T", bound=npt.ArrayLike)


def gaussian_pdf(x: T, mu: float = 0.0, sigma: float = 1.0) -> T:
	"""
	Computes the pdf of a Normal distribution.
	On my machine, is ~8x faster than scipy.stats.norm.pdf.

	@param x: ArrayLike
		The number or array on which to compute the pdf.
	@param mu: float
		The mean of the Normal distribution.
	@param sigma: float
	 	The standard deviation of the Normal distribution.
	@return gaussian_pdf: ArrayLike
		The computed pdf.
	"""

	return 1/(sigma * np.sqrt(2*np.pi)) * np.exp(-(x - mu)**2 / (2 * sigma**2))


def filter_kwargs(kwargs: dict[str, Any], function: Callable) -> dict[str, Any]:
	"""
	Filters the kwargs to only keep the keys that are accepted by the function.

	@param kwargs: dict[str, Any]
		The kwargs to filter.
	@param function: Callable
		The function receiving the arguments.
	@return filtered_kwargs: dict[str, Any]
		The filtered kwargs.
	"""

	signature = inspect.signature(function)
	filter_keys = [param.name for param in signature.parameters.values() if param.kind == param.POSITIONAL_OR_KEYWORD]

	return {filter_key: kwargs[filter_key] for filter_key in filter_keys if filter_key in kwargs}


def flatten_dict(d: dict, sep: str = ':', parent_key: str = '') -> dict:
	"""
	Flattens a nested dictionary.

	@param d: dict
	@param sep: str
		The separator between keys.
		The dictionary to flatten.
	@param parent_key: str
		The parent key (leave empty, used for recursion).
	@return flattened_dict: dict
		The flattened dictionary.
	"""

	items = []
	for k, v in d.items():
		new_key = parent_key + sep + k if parent_key else k
		if v and isinstance(v, dict):
			items.extend(flatten_dict(v, sep, parent_key=new_key).items())
		else:
			items.append((new_key, v))
	return dict(items)


def unflatten_dict(d: dict, sep: str = ':', base: dict | None = None) -> dict:
	"""
	Unflattens a dictionary.

	@param d: dict
		The dictionary to unflatten.
	@param sep: str
		The separator between keys.
	@param base: dict | None
		The base dictionary (leave at None, used for recursion).
	@return unflattened_dict: dict
		The unflattened dictionary.
	"""

	if base is None:
		base = {}

	for key, value in d.items():
		root = base

		if sep in key:
			*parts, key = key.split(sep)

			for part in parts:
				root.setdefault(part, {})
				root = root[part]

		if isinstance(value, dict):
			value = unflatten_dict(value, base=root.get(key, {}))

		root[key] = value

	return base


def merge_dict(d1: dict, d2: dict) -> dict:
	"""
	Merges two dictionaries (even in nested).
	If a key is present in both dictionaries, the value of the first dictionary is kept.

	@param d1: dict
		The first dictionary (which takes priority).
	@param d2: dict
		The second dictionary.
	@return merged_dict: dict
		The merged dictionary.
	"""

	res = {}

	for key in list(dict.fromkeys(([*d1.keys(), *d2.keys()]))):  # Important to keep the order of keys!
		if key not in d2:
			res[key] = d1[key]
		elif key not in d1:
			res[key] = d2[key]
		elif isinstance(d1[key], dict) and isinstance(d2[key], dict):
			res[key] = merge_dict(d1[key], d2[key])
		else:
			res[key] = d1[key]

	return res


def binom_sf(x: int, n: float, p: float) -> float:
	"""
	Computes the survival function (sf = 1 - cdf) of the binomial distribution.
	From values where the cdf is really close to 1.0, the survival function gives more precise results.
	Allows for a non-integer n (uses interpolation).

	@param x: int
		The number of successes.
	@param n: float
		The number of trials.
	@param p: float
		The probability of success.
	@return sf: float
		The survival function of the binomial distribution.
	"""

	n_array = np.arange(math.floor(n-2), math.ceil(n+3), 1)
	n_array = n_array[n_array >= 0]

	res = [scipy.stats.binom.sf(x, n_, p) for n_ in n_array]
	f = scipy.interpolate.interp1d(n_array, res, kind="quadratic")

	return f(n)


def gaussian_histogram(events: np.ndarray, t_axis: np.ndarray, sigma: float, truncate: float = 5., margin_reflect: bool = False) -> npt.NDArray[np.float32]:
	"""
	Computes a gaussian histogram for the given events.
	For each point in time, take all the nearby events and compute the sum of their gaussian kernel.

	@param events: np.ndarray
		The events to histogram.
	@param t_axis: np.ndarray (n_timepoints)
		The time axis of the histogram.
	@param sigma: float
		The standard deviation of the gaussian kernel (same unit as 't_axis').
	@param truncate: float
		Truncate the gaussian kernel at 'truncate' standard deviation.
	@param margin_reflect: bool
		If true, will reflect the events at the margins.
	@return histogram: np.ndarray[float32] (n_timepoints)
		The histogram of the events.
	"""

	if len(events) == 0:
		return np.zeros(len(t_axis), dtype=np.float32)

	events = np.sort(events).astype(np.float32)
	t_axis = t_axis.astype(np.float32)

	if margin_reflect:
		if np.min(events) >= t_axis[0]:
			events_low = 2*t_axis[0] - events[:np.searchsorted(events, t_axis[0] + truncate * sigma, side="right")][::-1]
		else:
			events_low = np.array([], dtype=np.float32)

		if np.max(events) <= t_axis[-1]:
			events_high = 2*t_axis[-1] - events[np.searchsorted(events, t_axis[-1] - truncate * sigma, side="left"):][::-1]
		else:
			events_high = np.array([], dtype=np.float32)

		events = np.concatenate((events_low, events, events_high))

	return _gaussian_kernel(events, t_axis, sigma, truncate)


@numba.jit((numba.float32[:], numba.float32[:], numba.float32, numba.float32), nopython=True, nogil=True, cache=True, parallel=True)
def _gaussian_kernel(events, t_axis, sigma, truncate) -> npt.NDArray[np.float32]:
	"""
	Numba function for gaussian_histogram.

	@param events: array[float32] (n_events)
		The ordered events timings.
	@param t_axis: array[float32] (n_timepoints)
		The time axis of the histogram.
	@param sigma: float32
		The standard deviation of the gaussian kernel (same unit as 't_axis').
	@param truncate: float32
		Truncate the gaussian kernel at 'truncate' standard deviation.
	@return histogram: array[float32] (n_timepoints)
		The gaussian histogram of the events.
	"""

	histogram = np.zeros(t_axis.shape, dtype=np.float32)

	start_j = 0
	for i in numba.prange(len(t_axis)):
		for e in events[start_j:]:
			if e < t_axis[i] - truncate * sigma:
				start_j += 1
				continue
			if e > t_axis[i] + truncate * sigma:
				break

			histogram[i] += np.exp(-0.5 * ((e - t_axis[i]) / sigma) ** 2)

	return histogram / (sigma * np.sqrt(2*np.pi))


@numba.jit((numba.int64[:], numba.int64[:]), nopython=True, nogil=True, cache=True)
def spike_vector_to_spike_trains(sample_indices, unit_indices) -> list[np.ndarray[np.int64]]:
	"""
	Converts a spike vector to a list of spike trains in a really fast manner.

	@param sample_indices: array[int64] (n_spikes1)
		All the spike timings.
	@param unit_indices: array[int64] (n_spikes1)
		The unit labels (i.e. unit index of each spike).
	@return spike_trains: list[array[int64]]
		The list of spike trains.
	"""

	num_units = (1 + np.max(unit_indices)) if len(unit_indices) > 0 else 0
	num_spikes = sample_indices.size

	num_spikes_per_unit = np.zeros(num_units, dtype=np.int32)
	for s in range(num_spikes):
		num_spikes_per_unit[unit_indices[s]] += 1

	spike_trains = []
	for u in range(num_units):
		spike_trains.append(np.empty(num_spikes_per_unit[u], dtype=np.int64))

	current_x = np.zeros(num_units, dtype=np.int32)
	for s in range(num_spikes):
		unit_index = unit_indices[s]
		spike_trains[unit_index][current_x[unit_index]] = sample_indices[s]
		current_x[unit_index] += 1

	return spike_trains


def estimate_contamination(spike_train: np.ndarray, refractory_period: tuple[float, float]) -> float:
	"""
	Estimates the contamination of a spike train by looking at the number of refractory period violations.
	The spike train is assumed to have spikes coming from a neuron, and noisy spikes that are random and
	uncorrelated to the neuron. Under this assumption, we can estimate the contamination (i.e. the
	fraction of noisy spikes to the total number of spikes).

	@param spike_train: np.ndarray
		The unit's spike train.
	@param refractory_period: tuple[float, float]
		The censored and refractory period (t_c, t_r) used (in ms).
	@return estimated_contamination: float
		The estimated contamination between 0 and 1.
	"""

	t_c = refractory_period[0] * 1e-3 * Utils.sampling_frequency
	t_r = refractory_period[1] * 1e-3 * Utils.sampling_frequency
	n_v = compute_nb_violations(spike_train.astype(np.int64), t_r)

	N = len(spike_train)
	D = 1 - n_v * (Utils.t_max - 2*N*t_c) / (N**2 * (t_r - t_c))
	contamination = 1.0 if D < 0 else 1 - math.sqrt(D)

	return contamination


def estimate_cross_contamination(spike_train1: np.ndarray, spike_train2: np.ndarray,
								 refractory_period: tuple[float, float], limit: float | None = None) -> tuple[float, float] | float:
	"""
	Estimates the cross-contamination of the second spike train with the neuron of the first spike train.
	Also performs a statistical test to check if the cross-contamination is significantly higher than a given limit.

	@param spike_train1: np.ndarray
		The spike train of the first unit.
	@param spike_train2: np.ndarray
		The spike train of the second unit.
	@param refractory_period: tuple[float, float]
		The censored and refractory period (t_c, t_r) used (in ms).
	@param limit: float | None
		The higher limit of cross-contamination for the statistical test.
	@return (estimated_cross_cont, p_value): tuple[float, float] if limit is not None
			estimated_cross_cont: float if limit is None
		Returns the estimation of cross-contamination, as well as the p-value of the statistical test if the limit is given.
	"""
	spike_train1 = spike_train1.astype(np.int64, copy=False)
	spike_train2 = spike_train2.astype(np.int64, copy=False)

	N1 = len(spike_train1)
	N2 = len(spike_train2)
	C1 = estimate_contamination(spike_train1, refractory_period)

	t_c = refractory_period[0] * 1e-3 * Utils.sampling_frequency
	t_r = refractory_period[1] * 1e-3 * Utils.sampling_frequency
	n_violations = compute_nb_coincidence(spike_train1, spike_train2, t_r) - compute_nb_coincidence(spike_train1, spike_train2, t_c)

	estimation = 1 - ((n_violations * Utils.t_max) / (2*N1*N2 * t_r) - 1) / (C1 - 1) if C1 != 1.0 else -np.inf
	if limit is None:
		return estimation

	# n and p for the binomial law for the number of coincidence (under the hypothesis of cross-contamination = limit).
	n = N1 * N2 * ((1 - C1) * limit + C1)
	p = 2 * t_r / Utils.t_max
	p_value = binom_sf(int(n_violations - 1), n, p)
	if np.isnan(p_value):  # pragma: no cover (should be unreachable).
		raise ValueError(f"Could not compute p-value for cross-contamination:\n\tn_violations = {n_violations}\n\tn = {n}\n\tp = {p}")

	return estimation, p_value


@numba.jit((numba.float32, ), nopython=True, nogil=True, cache=True)
def _get_border_probabilities(max_time) -> tuple[int, int, float, float]:
	"""
	Computes the integer borders, and the probability of 2 spikes distant by this border to be closer than max_time.

	@param max_time: float
		The maximum time between 2 spikes to be considered as a coincidence.
	@return border_low, border_high, p_low, p_high: tuple[int, int, float, float]
		The borders and their probabilities.
	"""

	border_high = math.ceil(max_time)
	border_low = math.floor(max_time)
	p_high = .5 * (max_time - border_high + 1) ** 2
	p_low  = .5 * (1 - (max_time - border_low)**2) + (max_time - border_low)

	if border_low == 0:
		p_low -= .5 * (-max_time + 1)**2

	return border_low, border_high, p_low, p_high


@numba.jit((numba.int64[:], numba.float32), nopython=True, nogil=True, cache=True)
def compute_nb_violations(spike_train, max_time) -> float:
	"""
	Computes the number of refractory period violations in a spike train.

	@param spike_train: array[int64] (n_spikes)
		The spike train to compute the number of violations for.
	@param max_time: float32
		The maximum time to consider for violations (in number of samples).
	@return n_violations: float
		The number of spike pairs that violate the refractory period.
	"""

	if max_time <= 0.0:
		return 0.0

	border_low, border_high, p_low, p_high = _get_border_probabilities(max_time)
	n_violations = 0
	n_violations_low = 0
	n_violations_high = 0

	for i in range(len(spike_train)-1):
		for j in range(i+1, len(spike_train)):
			diff = spike_train[j] - spike_train[i]

			if diff > border_high:
				break
			if diff == border_high:
				n_violations_high += 1
			elif diff == border_low:
				n_violations_low += 1
			else:
				n_violations += 1

	return n_violations + p_high*n_violations_high + p_low*n_violations_low


@numba.jit((numba.int64[:], numba.int64[:], numba.float32), nopython=True, nogil=True, cache=True)
def compute_nb_coincidence(spike_train1, spike_train2, max_time) -> float:
	"""
	Computes the number of coincident spikes between two spike trains.
	Spike timings are integers, so their real timing follows a uniform distribution between t - dt/2 and t + dt/2.
	Under the assumption that the uniform distributions from two spikes are independent, we can compute the probability
	of those two spikes being closer than the coincidence window:
	f(x) = 1/2 (x+1)² if -1 <= x <= 0
	f(x) = 1/2 (1-x²) + x if 0 <= x <= 1
	where x is the distance between max_time floor/ceil(max_time)

	@param spike_train1: array[int64] (n_spikes1)
		The spike train of the first unit.
	@param spike_train2: array[int64] (n_spikes2)
		The spike train of the second unit.
	@param max_time: float32
		The maximum time to consider for coincidence (in number samples).
	@return n_coincidence: float
		The number of coincident spikes.
	"""

	if max_time <= 0:
		return 0.0

	border_low, border_high, p_low, p_high = _get_border_probabilities(max_time)
	n_coincident = 0
	n_coincident_low = 0
	n_coincident_high = 0

	start_j = 0
	for i in range(len(spike_train1)):
		for j in range(start_j, len(spike_train2)):
			diff = spike_train1[i] - spike_train2[j]

			if diff > border_high:
				start_j += 1
				continue
			if diff < -border_high:
				break
			if abs(diff) == border_high:
				n_coincident_high += 1
			elif abs(diff) == border_low:
				n_coincident_low += 1
			else:
				n_coincident += 1

	return n_coincident + p_high*n_coincident_high + p_low*n_coincident_low


def compute_coincidence_matrix_from_vector(spike_vector1: np.ndarray, spike_vector2: np.ndarray, window: int, cross_shifts: np.ndarray | None = None) -> npt.NDArray[np.int64]:
	"""
	Computes the number of coincident spikes between two sortings (given their spike vector).

	@param spike_vector1: np.ndarray (n_spikes1)
		The spike vector of the first sorting.
	@param spike_vector2: np.ndarray (n_spikes2)
		The spike vector of the second sorting.
	@param window: int
		The coincidence window (in number of samples).
		Two spikes separated by exactly window are considered as coincident.
	@param cross_shifts: None | array[int32] (n_units1, n_units2)
		If not None, the cross_shifts[i, j] is the shift between the spike times of the i-th unit of the first sorting
		and the j-th unit of the second sorting.
	@return coincidence_matrix: np.ndarray[int64] (n_units1, n_units2)
		The coincidence matrix containing the number of coincident spikes between each pair of units.
	"""

	if cross_shifts is not None:
		cross_shifts = cross_shifts.astype(np.int32)

	return compute_coincidence_matrix(spike_vector1['sample_index'], spike_vector1['unit_index'],
									  spike_vector2['sample_index'], spike_vector2['unit_index'], window, cross_shifts)


@numba.jit((numba.int64[:], numba.int64[:], numba.int64[:], numba.int64[:], numba.int32, numba.optional(numba.int32[:, :])),
		   nopython=True, nogil=True, cache=True)
def compute_coincidence_matrix(spike_times1, spike_labels1, spike_times2, spike_labels2, max_time, cross_shifts=None) -> npt.NDArray[np.int64]:
	"""
	Computes the number of coincident spikes between all units in two sortings.

	@param spike_times1: array[int64] (n_spikes1)
		All the spike timings of the first sorting.
	@param spike_labels1: array[int64] (n_spikes1)
		The unit labels of the first sorting (i.e. unit index of each spike).
	@param spike_times2: array[int64] (n_spikes2)
		All the spike timings of the second sorting.
	@param spike_labels2: array[int64] (n_spikes2)
		The unit labels of the second sorting (i.e. unit index of each spike).
	@param max_time: int32
		The maximum time difference between two spikes to be considered coincident.
		Two spikes spaced by exactly max_time are considered coincident.
	@param cross_shifts: None | array[int32] (n_units1, n_units2)
		If not None, the cross_shifts[i, j] is the shift between the spike times of the i-th unit of the first sorting
		and the j-th unit of the second sorting.
	@return coincidence_matrix: array[int64] (n_units1, n_units2)
		The coincidence matrix containing the number of coincident spikes between each pair of units.
	"""

	if cross_shifts is None:
		n_units1 = (np.max(spike_labels1) + 1) if len(spike_labels1) > 0 else 0
		n_units2 = (np.max(spike_labels2) + 1) if len(spike_labels2) > 0 else 0
		cross_shifts = np.zeros((n_units1, n_units2), dtype=np.int32)

	coincidence_matrix = np.zeros(cross_shifts.shape, dtype=np.int64)

	start_j = 0
	for i in range(len(spike_times1)):
		for j in range(start_j, len(spike_times2)):
			diff = spike_times1[i] - spike_times2[j] - cross_shifts[spike_labels1[i], spike_labels2[j]]

			if diff > max_time:
				start_j += 1
				continue
			if diff < -max_time:
				break

			coincidence_matrix[spike_labels1[i], spike_labels2[j]] += 1

	return coincidence_matrix


def compute_similarity_matrix(coincidence_matrix: np.ndarray, n_spikes1: np.ndarray, n_spikes2: np.ndarray, window: float = -.5):
	"""
	Computes the similarity matrix from the coincidence matrix.

	@param coincidence_matrix: array[int] (n_units1, n_units2)
		The coincidence matrix between the two sortings.
	@param n_spikes1: array[int] (n_units1)
		The number of spikes for each unit in the first sorting.
	@param n_spikes2: array[int] (n_units2)
		The number of spikes for each unit in the second sorting.
	@param window: int
		The window used for the coincidence matrix (to compute the corrected similarity matrix).
		Leave at -0.5 to compute the uncorrected similarity matrix.
	@return similarity_matrix: array[float] (n_units1, n_units2)
		The similarity matrix between the two sortings.
	"""
	assert coincidence_matrix.shape == (len(n_spikes1), len(n_spikes2))

	minimum_n_spikes = np.minimum(n_spikes1[:, None], n_spikes2)
	similarity_matrix = coincidence_matrix / minimum_n_spikes
	expected_matrix = (n_spikes1[:, None].astype(np.int64) * n_spikes2[None, :].astype(np.int64) * (2*window+1) / Utils.t_max) / minimum_n_spikes

	return (similarity_matrix - expected_matrix) / (1 - expected_matrix)


def compute_cross_shift_from_vector(spike_vector1: np.ndarray, spike_vector2: np.ndarray, max_shift: int, gaussian_std: float = 1.5) -> npt.NDArray[np.int32]:
	"""
	Computes the shift between units pairwise between 2 sortings (given their spike vector).
	Looks at their spike times and creates a cross-correlogram to look for a central peak.

	@param spike_vector1: np.ndarray (n_spikes1)
		The spike vector of the first sorting.
	@param spike_vector2: np.ndarray (n_spikes2)
		The spike vector of the second sorting.
	@param max_shift: int32
		The maximum shift to consider (in samples).
	@param gaussian_std: float32
		The standard deviation of the Gaussian kernel used to smooth the cross-correlogram.
	@return cross_shift_matrix: array[int32] (n_units1, n_units2)
		The cross-shift matrix containing the shift between each pair of units.
	"""

	return compute_cross_shift(spike_vector1['sample_index'].astype(np.int64, copy=False), spike_vector1['unit_index'].astype(np.int64, copy=False),
							   spike_vector2['sample_index'].astype(np.int64, copy=False), spike_vector2['unit_index'].astype(np.int64, copy=False), max_shift, gaussian_std)


@numba.jit((numba.int64[:], numba.int64[:], numba.int64[:], numba.int64[:], numba.int32, numba.float32),
		   nopython=True, nogil=True, cache=True, parallel=True)
def compute_cross_shift(spike_times1, spike_labels1, spike_times2, spike_labels2, max_shift, gaussian_std) -> npt.NDArray[np.int32]:
	"""
	Computes the shift between units pairwise between 2 sortings.
	Looks at their spike times and creates a cross-correlogram to look for a central peak.

	@param spike_times1: array[int64] (n_spikes1)
		All the spike timings of the first sorting.
	@param spike_labels1: array[int64] (n_spikes1)
		The unit labels of the first sorting (i.e. unit index of each spike).
	@param spike_times2: array[int64] (n_spikes2)
		All the spike timings of the second sorting.
	@param spike_labels2: array[int64] (n_spikes2)
		The unit labels of the second sorting (i.e. unit index of each spike).
	@param max_shift: int32
		The maximum shift to consider (in samples).
	@param gaussian_std: float32
		The standard deviation of the Gaussian kernel used to smooth the cross-correlogram.
	@return cross_shift_matrix: array[int32] (n_units1, n_units2)
		The cross-shift matrix containing the shift between each pair of units.
	"""

	n_units1 = (np.max(spike_labels1) + 1) if len(spike_labels1) > 0 else 0
	n_units2 = (np.max(spike_labels2) + 1) if len(spike_labels2) > 0 else 0
	cross_shift_matrix = np.zeros((n_units1, n_units2), dtype=np.int32)

	N = math.ceil(5 * gaussian_std)
	gaussian = np.exp(-np.arange(-N, N+1)**2 / (2 * gaussian_std**2)) / (gaussian_std * math.sqrt(2*math.pi))

	spike_trains1 = spike_vector_to_spike_trains(spike_times1, spike_labels1)
	spike_trains2 = spike_vector_to_spike_trains(spike_times2, spike_labels2)

	for unit1 in numba.prange(n_units1):
		for unit2 in range(n_units2):
			spike_train1 = spike_trains1[unit1]
			spike_train2 = spike_trains2[unit2]
			threshold = 0.1 * min(len(spike_train1), len(spike_train2))

			cross_corr = _compute_crosscorr_numba(spike_train1, spike_train2, max_shift, 1)
			cross_corr = np.convolve(cross_corr, gaussian)
			idx = np.argmax(cross_corr)

			if np.sum(cross_corr[idx-1:idx+2]) < threshold:  # TODO: Better way than idx-1 : idx+2
				continue

			cross_shift_matrix[unit1, unit2] = idx - len(cross_corr) // 2

	return cross_shift_matrix


def filter(data: np.ndarray, band: tuple[float, float] | list[float, float] | np.ndarray, axis: int = -1) -> np.ndarray:
	"""
	Filters the data using a Gaussian bandpass filter.
	Since the Gaussian filter is a convolution with a Gaussian kernel, we can accelerate the filtering
	by using FFTs. The convolution in the time domain is a multiplication in the frequency domains.

	@param data: np.ndarray
		The data to filter.
	@param band: Iterable of two floats [highpass, lowpass]
		The highpass and lowpass frequencies (in Hz).
	@param axis: int
		The axis along which to filter the data.
	@return filtered_data: np.ndarray
		The filtered data.
	"""

	N = data.shape[axis]
	data_fft = np.fft.fft(data, axis=axis)
	gaussian_highpass = _create_fft_gaussian(N, band[0])
	gaussian_lowpass  = _create_fft_gaussian(N, band[1])

	broadcast = tuple(slice(None) if i == axis else None for i in range(data_fft.ndim))
	filtered_data_fft = data_fft * (gaussian_lowpass - gaussian_highpass)[broadcast]
	filtered_data = np.fft.ifft(filtered_data_fft, axis=axis).real

	return filtered_data


def _create_fft_gaussian(N: int, cutoff_freq: float) -> np.ndarray:
	"""
	Creates a Gaussian filter in the frequency domain for a given cutoff frequency.

	@param N: int
		The number of samples of the signal for the Fourier transform.
	@param cutoff_freq: float
		The cutoff frequency (in Hz).
	@return fft_gaussian: np.ndarray
		The Gaussian filter in the frequency domain.
	"""

	if cutoff_freq > Utils.sampling_frequency / 8:  # The Fourier transform of a Gaussian with a very low sigma isn't a Gaussian in the Fourier domain.
		sigma = Utils.sampling_frequency / (2 * math.pi * cutoff_freq)
		limit = int(round(6*sigma)) + 1
		xaxis = np.arange(-limit, limit+1) / sigma
		gaussian = gaussian_pdf(xaxis) / sigma
		return np.abs(np.fft.fft(gaussian, n=N))
	else:
		freq_axis = np.fft.fftfreq(N, d=1/Utils.sampling_frequency)
		return gaussian_pdf(freq_axis / cutoff_freq) * math.sqrt(2 * math.pi)


def compute_correlogram_difference(auto_corr1: np.ndarray, auto_corr2: np.ndarray, cross_corr: np.ndarray, n1: int, n2: int) -> float:
	"""
	Code to compute the correlogram difference between two units.
	The idea is to compare both auto-correlograms to the cross-correlogram,
	weighted by the number of spikes in each unit (the unit with more spikes imposes its result).

	@param auto_corr1: np.ndarray
		The auto-correlogram of the first unit.
	@param auto_corr2: np.ndarray
		The auto-correlogram of the second unit.
	@param cross_corr: np.ndarray
		The cross-correlogram between the two units.
	@param n1: int
		The number of spikes in the first unit.
	@param n2: int
		The number of spikes in the second unit.
	@return difference: float
		The computed correlogram difference between both units (0.0 = they are similar).
	"""

	auto_corr1 = normalize_correlogram(auto_corr1)
	auto_corr2 = normalize_correlogram(auto_corr2)
	cross_corr = normalize_correlogram(cross_corr)

	# Windows
	middle = len(auto_corr1) // 2
	w1 = get_unit_adaptive_window(auto_corr1, 0.5)
	w2 = get_unit_adaptive_window(auto_corr2, 0.5)
	w = int(round((w1*n1 + w2*n2) / (n1 + n2)))
	window = slice(middle - w, middle + w + 1)

	diff1 = np.sum(np.abs(cross_corr[window] - auto_corr1[window])) / (window.stop - window.start)
	diff2 = np.sum(np.abs(cross_corr[window] - auto_corr2[window])) / (window.stop - window.start)
	weighted_diff = (n1*diff1 + n2*diff2) / (n1+n2)

	return weighted_diff



from .misc import (gaussian_pdf, filter_kwargs, flatten_dict, unflatten_dict, merge_dict, binom_sf, gaussian_histogram, estimate_contamination, estimate_cross_contamination, compute_nb_coincidence,
				   compute_coincidence_matrix, compute_coincidence_matrix_from_vector, compute_similarity_matrix, filter, compute_cross_shift_from_vector,
				   compute_cross_shift, compute_correlogram_difference)
from .plotting import get_path_to_plotlyJS, export_figure, plot_sliders, plot_units, create_gt_annotations, create_graph_plot
from .variables import Utils



from typing import Any
from overrides import override
from lussac.core import MonoSortingModule
import spikeinterface.core as si
import spikeinterface.curation as scur


class RemoveDuplicatedSpikes(MonoSortingModule):
	"""
	Removes duplicated spikes from all units.
	Spikes are considered duplicated if they are separated by less than x time samples.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'censored_period': 0.3,
			'method': "keep_first_iterative"
		}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		return scur.remove_duplicated_spikes(self.sorting, params['censored_period'], params['method'])



from .lussac_data import LussacData, MonoSortingData, MultiSortingsData
from .module import LussacModule, MonoSortingModule, MultiSortingsModule
from .module_factory import ModuleFactory
from .pipeline import LussacPipeline
from .spike_sorting import LussacSpikeSorter



from typing import Any
import numpy as np
from overrides import override
from lussac.core import MonoSortingModule
import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.curation as scur


class RemoveRedundantUnits(MonoSortingModule):
	"""
	Removes units that are redundant with other units in the same sorting
	(i.e. they share similar spike timings over a certain threshold).
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'wvf_extraction': {
				'ms_before': 1.0,
				'ms_after': 1.5,
				'max_spikes_per_unit': 500
			},
			'arguments': {
				'align': True,
				'delta_time': 0.3,
				'agreement_threshold': 0.1,
				'duplicate_threshold': 0.7,
				'remove_strategy': 'highest_amplitude'
			}
		}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		sorting_or_wvf_extractor = self.extract_waveforms(**params['wvf_extraction']) if params['wvf_extraction'] is not None else self.sorting

		new_sorting, redundant_unit_pairs = scur.remove_redundant_units(sorting_or_wvf_extractor, extra_outputs=True, **params['arguments'])

		redundant_unit_ids = [unit_id for unit_id in self.sorting.unit_ids if unit_id not in new_sorting.unit_ids]
		redundant_sorting = self.sorting.select_units(redundant_unit_ids)
		redundancies = self._get_redundancies(redundant_unit_ids, redundant_unit_pairs)
		self._plot_redundant_units(redundant_sorting, redundancies)

		return new_sorting

	@staticmethod
	def _get_redundancies(redundant_unit_ids: list, redundant_unit_pairs: list[list]) -> dict:
		"""
		For each redundant units, finds which units it is redundant with.

		@param redundant_unit_ids: list
			A list of all redundant units' id.
		@param redundant_unit_pairs: list[list]
			A list of pairs of redundant units.
		@return redundancies: dict
			For each redundant unit (key = id), a list of the units it is redundant with (value = list of ids).
		"""

		redundancies = {}
		for unit_id in redundant_unit_ids:
			redundant_with = []
			for pair in redundant_unit_pairs:
				if unit_id in pair:
					redundant_with.append(pair[0] if unit_id == pair[1] else pair[1])
			redundancies[unit_id] = redundant_with

		return redundancies

	def _plot_redundant_units(self, redundant_sorting: si.BaseSorting, redundancies: dict) -> None:
		"""
		Plots the units that were removed.

		@param redundant_sorting: si.BaseSorting
			The sorting object containing the redundant units.
		"""

		wvf_extractor = self.extract_waveforms(sorting=redundant_sorting, sub_folder="plot_redundant", ms_before=1.5, ms_after=2.5, max_spikes_per_unit=500)
		annotations = [{'text': f"Unit {unit_id} is redundant with unit(s): {' '.join(np.array(redundancies[unit_id]).astype(str))}", 'x': 0.6, 'y': 1.07,
						'xref': "paper", 'yref': "paper", 'showarrow': False} for unit_id in redundant_sorting.unit_ids]
		utils.plot_units(wvf_extractor, filepath=f"{self.logs_folder}/redundant_units", annotations_change=annotations)



from typing import Any
import numpy as np
from overrides import override
from lussac.core import MonoSortingModule
import spikeinterface.core as si
import spikeinterface.postprocessing as spost
import spikeinterface.qualitymetrics as sqm


class FindPurkinjeCells(MonoSortingModule):
	"""
	This module is meant purely for the cerebellar cortex.
	Purkinje cells are big GABAergic neurons that can fire both simple spikes and complex spikes.
	Since simple spikes and complex spikes have different shape, they will be clustered in different units.
	This module will find simple spikes and complex spikes coming from the same Purkinje cell,
	and label then into a 'lussac_purkinje' property in the sorting object.
	This is done by looking at the cross-correlogram between the two units: after a complex spike,
	there is always a pause of at least 8-10 ms before simple spike activity resumes.
	Also, when high-pass filtered aggressively, the templates are nearly identical.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'cross_corr_pause': [0.0, 8.0],
			'threshold': 0.4,
			'ss_min_fr': 40.0,
			'cs_min_fr': 0.5,
			'cs_max_fr': 3.0
		}

	@override
	def update_params(self, params: dict[str, Any]) -> dict[str, Any]:
		params = super().update_params(params)
		if params['ss_min_fr'] <= params['cs_max_fr']:
			raise ValueError("Error in 'find_purkinje_cells': 'ss_min_fr' must be greater than 'cs_max_fr'")
		if params['threshold'] < 0.0:
			raise ValueError("Error in 'find_purkinje_cells': 'threshold' cannot be negative")

		return params

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		wvf_extractor = si.WaveformExtractor(self.recording, self.sorting, allow_unfiltered=True)
		firing_rates = sqm.compute_firing_rates(wvf_extractor)

		putative_ss_units = [unit_id for unit_id, mean_fr in firing_rates.items() if mean_fr >= params['ss_min_fr']]
		putative_cs_units = [unit_id for unit_id, mean_fr in firing_rates.items() if params['cs_min_fr'] <= mean_fr <= params['cs_max_fr']]

		sorting = self.sorting.select_units([*putative_ss_units, *putative_cs_units])
		correlograms, bins = spost.compute_correlograms(sorting, window_ms=25.0, bin_ms=1.0, method="numba")
		mask = ((bins >= params['cross_corr_pause'][0]) & (bins < params['cross_corr_pause'][1]))[:-1]

		ss_cs_pairs = []
		for ss_id in putative_ss_units:
			for cs_id in putative_cs_units:
				ss_ind = sorting.id_to_index(ss_id)
				cs_ind = sorting.id_to_index(cs_id)
				cross_corr = correlograms[ss_ind, cs_ind, :]

				baseline = np.median(cross_corr[bins[:-1] < 0.0])
				if np.median(cross_corr[mask]) < baseline * params['threshold']:  # Check for pause.
					if np.median(cross_corr[mask[::-1]] < baseline * params['threshold']):  # Check for asymmetry.
						continue
					ss_cs_pairs.append((ss_id, cs_id))  # TODO: Also check templates?

		lussac_purkinje = np.empty(self.sorting.get_num_units(), dtype=object)
		paired_units = np.unique(np.array(ss_cs_pairs).flatten())
		for unit_id in paired_units:
			unit_ind = self.sorting.id_to_index(unit_id)
			mask = np.array([unit_id in ss_cs_pairs[i] for i in range(len(ss_cs_pairs))])
			text = " ; ".join([f"{ss_id}-{cs_id}" for ss_id, cs_id in np.array(ss_cs_pairs)[mask]])
			lussac_purkinje[unit_ind] = text

		self.sorting.set_property('lussac_purkinje', lussac_purkinje)
		return self.sorting



import pathlib
from typing import ClassVar


class Utils:
	"""
	Class containing static variables that are useful for the whole package.

	Variables:
		sampling_frequency	The sampling frequency of the recording (in Hz).
		t_max				The number of time samples in the recording.
		plotly_js_file		The path to the plotly.js file.

	"""

	sampling_frequency:	ClassVar[int]
	t_max:				ClassVar[int]
	plotly_js_file:		ClassVar[pathlib.Path]



import os
import requests
import pathlib
import pytest
import shutil
from tqdm import tqdm
import lussac.main
from conftest import params_path
import spikeinterface.core as si


def test_dataset_exists(capsys):
	if not params_path.exists():
		with capsys.disabled():  # Don't capture the print output and show it.
			file_path = pathlib.Path(__file__).parent / "datasets" / "cerebellar_cortex.zip"
			file_path.parent.mkdir(parents=True, exist_ok=True)

			http_response = requests.get("https://zenodo.org/record/8171929/files/lussac2_cerebellar_cortex_dev.zip", stream=True)
			n_bytes = int(http_response.headers.get("content-length"))

			with tqdm.wrapattr(open(file_path, 'wb'), "write", miniters=1, desc=f"Downloading {file_path.name}", total=n_bytes) as fout:
				for chunk in http_response:
					fout.write(chunk)
				fout.close()  # Necessary even inside a 'with'.

			print("Unzipping ...")
			shutil.unpack_archive(file_path, file_path.parent)
			file_path.unlink()

			zarr_folder = file_path.parent / "cerebellar_cortex" / "recording.zarr"
			recording = si.ZarrRecordingExtractor(zarr_folder)
			recording.save(format="binary", folder=zarr_folder.parent / "recording.bin", n_jobs=2, chunk_duration='2s')

			print("")
			print(pathlib.Path(__file__).relative_to(pathlib.Path(os.getcwd())), end=' ')  # To have a nice output in the console.

	assert params_path.exists()


def test_parse_arguments() -> None:
	with pytest.raises(SystemExit):  # There is no arguments.
		lussac.main.parse_arguments(None)

	assert lussac.main.parse_arguments([str(params_path)]) == str(params_path)



from typing import Any
from overrides import override
import numpy as np
import numpy.typing as npt
from lussac.core import MonoSortingModule
import spikeinterface.core as si


class UnitsCategorization(MonoSortingModule):
	"""
	Categorizes units based on their properties.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		for category, rules in params.items():
			units_to_categorize = self._init_units_to_categorize()

			if category == "clear":
				self.sorting.set_property("lussac_category", None)
				continue

			for attribute, p in rules.items():
				value = self.get_units_attribute_arr(attribute, p)
				if 'min' in p:
					units_to_categorize &= value > p['min']
				if 'max' in p:
					units_to_categorize &= value < p['max']

			unit_ids = self.sorting.unit_ids[units_to_categorize]
			values = np.array([category]*len(unit_ids), dtype=object)  # dtype cannot be str, otherwise string length is restricted.
			self.sorting.set_property("lussac_category", values, ids=unit_ids, missing_value=None)

		return self.sorting

	def _init_units_to_categorize(self) -> npt.NDArray[np.integer]:
		"""
		Returns an array of True/False, where True means that the unit hasn't been categorized yet.

		@return units_to_categorize: np.ndarray[bool]
			Units that haven't been categorized yet.
		"""

		if self.sorting.get_property("lussac_category") is None:
			return np.ones(self.sorting.get_num_units(), dtype=bool)
		else:
			categories = self.sorting.get_property("lussac_category")
			return np.array([category is None for category in categories], dtype=bool)



from dataclasses import dataclass
import logging
import pathlib
import spikeinterface.core as si
import spikeinterface.curation as scur
import spikeinterface.preprocessing as spre
import spikeinterface.sorters as ss


@dataclass(slots=True)
class LussacSpikeSorter:
	"""
	Abstract class for all spike sorting algorithms running through Lussac.
	All children must have names to recognize them, and a launch method.

	Attributes:
		recording		A SpikeInterface recording object.
		name			The name of the analysis (not the sorting algorithm).
	"""

	recording: si.BaseRecording
	name: str

	def _preprocess(self, params: dict) -> None:
		"""
		Preprocesses the recording object based on what's in the params.

		@param params: dict
			Dictionary mapping the preprocessing functions to their arguments
		"""

		for preprocess_func, arguments in params.items():
			function = getattr(spre, preprocess_func)
			self.recording = function(self.recording, **arguments)

	def launch(self, params: dict) -> si.BaseSorting:
		"""
		Launches the spike sorting algorithm and returns the sorting object.

		@param params: dict
			The parameters for the spike sorting algorithm.
		"""

		sorter_name = params['sorter_name']
		folder = pathlib.Path(params['sorter_params']['output_folder'])

		if (folder / "provenance.pkl").exists():
			sorting = si.load_extractor(folder / "provenance.pkl")
			logging.info(f"Loading analysis '{self.name}'\n")
			assert isinstance(sorting, si.BaseSorting)
			return sorting

		if 'preprocessing' in params:
			self._preprocess(params['preprocessing'])

		logging.info(f"Running spike-sorting for analysis '{self.name}'\n")

		sorting = ss.run_sorter(sorter_name, self.recording, **params['sorter_params'])
		sorting = scur.remove_excess_spikes(sorting.remove_empty_units(), self.recording)
		sorting.annotate(name=self.name)
		sorting.dump_to_pickle(file_path=folder / "provenance.pkl", include_properties=True)

		return sorting



from typing import Any
from overrides import override
import numpy as np
from lussac.core import MonoSortingModule
import lussac.utils as utils
import spikeinterface.core as si


class RemoveBadUnits(MonoSortingModule):
	"""
	Removes the bad units from the sorting object.
	The definition of "bad" unit is given by the parameters dictionary.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		units_to_remove = np.zeros(self.sorting.get_num_units(), dtype=bool)
		reasons_for_removal = np.array([''] * self.sorting.get_num_units(), dtype=object)

		for attribute, p in params.items():
			if attribute == "all":
				units_to_remove[:] = True
				reasons_for_removal[:] += " ; all"
				break

			value = self.get_units_attribute_arr(attribute, p)
			if 'min' in p:
				units_to_remove |= value < p['min']
				reasons_for_removal[value < p['min']] += f" ; {attribute} < {p['min']}"
			if 'max' in p:
				units_to_remove |= value > p['max']
				reasons_for_removal[value > p['max']] += f" ; {attribute} > {p['max']}"

		sorting = self.sorting.select_units([unit_id for unit_id, bad in zip(self.sorting.unit_ids, units_to_remove) if not bad])
		bad_sorting = self.sorting.select_units([unit_id for unit_id, bad in zip(self.sorting.unit_ids, units_to_remove) if bad])
		reasons_for_removal = ["Reason(s) for removal: " + reason[3:] for reason in reasons_for_removal[units_to_remove]]

		self._plot_bad_units(bad_sorting, reasons_for_removal)

		return sorting

	def _plot_bad_units(self, bad_sorting: si.BaseSorting, reasons_for_removal: list[str]) -> None:
		"""
		Plots the units that were removed.

		@param bad_sorting: si.BaseSorting
			The sorting object containing the bad units.
		"""

		wvf_extractor = self.extract_waveforms(sorting=bad_sorting, ms_before=1.5, ms_after=2.5, max_spikes_per_unit=500, sparse=False)
		annotations = [{'text': reason, 'x': 0.6, 'y': 1.02, 'xref': "paper", 'yref': "paper", 'xanchor': "center", 'yanchor': "bottom", 'showarrow': False} for reason in reasons_for_removal]
		utils.plot_units(wvf_extractor, filepath=f"{self.logs_folder}/bad_units", annotations_change=annotations)



from .align_units import AlignUnits
from .export_to_phy import ExportToPhy
from .export_to_sigui import ExportToSIGUI
from .find_purkinje_cells import FindPurkinjeCells
from .merge_sortings import MergeSortings
from .merge_units import MergeUnits
from .remove_bad_units import RemoveBadUnits
from .remove_duplicated_spikes import RemoveDuplicatedSpikes
from .remove_redundant_units import RemoveRedundantUnits
from .units_categorization import UnitsCategorization



import copy
from abc import ABC, abstractmethod
from dataclasses import dataclass
import pathlib
import shutil
from typing import Any
import numpy as np
from lussac.core import MonoSortingData, MultiSortingsData
import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.preprocessing as spre
import spikeinterface.postprocessing as spost
import spikeinterface.qualitymetrics as sqm


@dataclass(slots=True)
class LussacModule(ABC):
	"""
	The abstract Module class.
	Every module used in Lussac must inherit from this class.

	Attributes:
		name			Module's name (i.e. the key in the pipeline dictionary).
		data 			Reference to the data object.
		category		What category is used for the module (i.e. the key in the dictionary).
		export_sortings	Whether to export the sortings after the module is executed.
	"""

	name: str
	data: MonoSortingData | MultiSortingsData
	category: str
	export_sortings: bool = True

	@property
	def recording(self) -> si.BaseRecording:
		"""
		Returns the recording object.

		@return recording: BaseRecording
			The recording object.
		"""

		return self.data.recording

	@property
	def sampling_f(self) -> float:
		"""
		Returns the sampling frequency of the recording (in Hz).

		@return sampling_f: float
			The sampling frequency (in Hz).
		"""

		return self.recording.sampling_frequency

	@abstractmethod
	def run(self, params: dict[str, Any]) -> si.BaseSorting | dict[str, si.BaseSorting]:
		"""
		Executes the module and returns the result (either a sorting of a dict of sortings).

		@param params: dict
			The parameters for the module.
		@return result: si.BaseSorting | dict[str, si.BaseSorting]
			The result of the module.
		"""
		...

	@property
	@abstractmethod
	def default_params(self) -> dict[str, Any]:
		"""
		Returns the default parameters of the module.

		@return default_params: dict[str, Any]
			The default parameters of the module.
		"""
		...

	def update_params(self, params: dict[str, Any]) -> dict[str, Any]:
		"""
		Updates the parameters with the default parameters (if there is a conflict, the given parameters are taken).
		Need to update recursively because the parameters are usually a nested dictionary.
		This is done by flattening then unflattening the dictionaries.

		@param params: dict[str, Any]
			The parameters to update.
		@return updated_params: dict[str, Any]
			The parameters updated with the module's default parameters.
		"""

		return utils.merge_dict(params, self.default_params)

	def extract_waveforms(self, sorting: si.BaseSorting, sub_folder: str | None = None, filter: list[float, float] | None = None, **params) -> si.WaveformExtractor:
		"""
		Creates the WaveformExtractor object and returns it.

		@param sorting: BaseSorting
			The sorting for the WaveformExtractor.
		@param sub_folder: str | None:
			The sub-folder where to save the waveforms.
		@param params
			The parameters for the waveform extractor.
		@param filter: list[float, float] | None
			The cutoff frequencies for the Gaussian bandpass filter to apply to the recording.
		@return wvf_extractor: WaveformExtractor
			The waveform extractor object.
		"""
		if sub_folder is None:
			sub_folder = "wvf_extractor"

		folder_path = f"{self.data.tmp_folder}/{self.name}/{self.category}/{sorting.get_annotation('name')}/{sub_folder}"

		recording = self.recording
		if filter is not None:
			assert len(filter) == 2, "The filter must be a list of 2 elements [min_cutoff, max_cutoff] (in Hz)."
			recording = spre.gaussian_filter(recording, *filter, margin_sd=2)

		sorting = sorting.to_numpy_sorting()  # Convert sorting for faster extraction.
		return si.extract_waveforms(recording, sorting, folder_path, allow_unfiltered=True, **params)


@dataclass(slots=True)
class MonoSortingModule(LussacModule):
	"""
	The abstract mono-sorting module class.
	This is for modules that don't work on multiple sortings at once.

	Attributes:
		name			Module's name (i.e. the key in the pipeline dictionary).
		data			Reference to the mono-sorting data object.
		category		What category is used for the module (i.e. the key in the dictionary).
		export_sortings	Whether to export the sortings after the module is executed.
	"""

	data: MonoSortingData

	def __del__(self) -> None:
		"""
		When the module is garbage collected, remove the temporary folder.
		"""

		if (self.data.tmp_folder / self.name).exists():
			shutil.rmtree(self.data.tmp_folder / self.name)

	@property
	def sorting(self) -> si.BaseSorting:
		"""
		Returns the sorting object.

		@return sorting: BaseSorting
			The sorting object.
		"""

		return self.data.sorting

	@property
	def logs_folder(self) -> pathlib.Path:
		"""
		Returns the logs directory for this module.

		@return logs_folder: Path
			Path to the logs directory.
		"""

		logs_folder = self.data.logs_folder / self.name / self.category / self.data.name
		logs_folder.mkdir(parents=True, exist_ok=True)

		return logs_folder

	@property
	def tmp_folder(self) -> pathlib.Path:
		"""
		Returns the temporary directory for this module.

		@return: tmp_folder: Path
			Path to the temporary directory.
		"""

		tmp_folder = self.data.tmp_folder / self.name / self.category / self.data.name
		tmp_folder.mkdir(parents=True, exist_ok=True)

		return tmp_folder

	@abstractmethod
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		...

	def extract_waveforms(self, sorting: si.BaseSorting | None = None, sub_folder: str | None = None, filter: list[float, float] | None = None, **params) -> si.WaveformExtractor:
		"""
		Calls the parent LussacModule.extract_waveforms
		'sorting' argument is optional. If None (default), will take the MonoSortingModule.data.sorting
		"""

		if sorting is None:
			sorting = self.sorting

		return super(MonoSortingModule, self).extract_waveforms(sorting, sub_folder, filter, **params)

	def get_templates(self, params: dict, filter_band: tuple[float, float] | list[float, float] | np.ndarray | None = None, margin: float = 3.0,
					  sub_folder: str = "templates", return_extractor: bool = False) -> np.ndarray | tuple[np.ndarray, si.WaveformExtractor, int]:
		"""
		Extract the templates for all the units.
		If filter_band is not None, will also filter them using a Gaussian filter.

		@param params: dict
			The parameters for the waveform extraction.
		@param filter_band: Iterable[float, float] | None
			If not none, the highpass and lowpass cutoff frequencies for Gaussian filtering (in Hz).
		@param margin: float
			The margin (in ms) to extract (useful for filtering).
		@param sub_folder: str
			The sub-folder used for the waveform extractor.
		@param return_extractor: bool
			If true, will also return the waveform extractor and margin (in samples).
		@return templates: np.ndarray (n_units, n_samples, n_channels)
			The extracted templates
		@return wvf_extractor: si.WaveformExtractor
			The Waveform Extractor of unfiltered waveforms.
			Only if return_extractor is True.
		@return margin: int
			The margin (in samples) that were used for the filtering.
			Only if return_extractor is True.
		"""

		params = params.copy()
		params['ms_before'] += margin
		params['ms_after'] += margin
		wvf_extractor = self.extract_waveforms(sub_folder=sub_folder, **params)
		templates = wvf_extractor.get_all_templates()

		if filter_band is not None:
			templates = utils.filter(templates, filter_band, axis=1)

		margin = int(round(margin * self.recording.sampling_frequency * 1e-3))

		if return_extractor:
			return templates[:, margin:-margin], wvf_extractor, margin
		else:
			return templates[:, margin:-margin]

	def get_units_attribute(self, attribute: str, params: dict) -> dict:
		"""
		Gets the attribute for all the units.

		@param attribute: str
			The attribute name.
			- firing_rate (in Hz)
			- contamination (between 0 and 1)
			- amplitude (unit depends on the wvf extractor 'return_scaled' parameter)
			- amplitude_std (unit depends on parameters 'return_scaled')
		@param params: dict
			The parameters to get the attribute.
			- 'filter': parameters to filter the recording.
			- 'wvf_extraction': parameters to extract the waveforms.
			- others: parameters for how to get the attribute.
		@return attribute: np.ndarray
			The attribute for all the units.
		"""
		params = copy.deepcopy(params)
		default_params = {
			'firing_rate': {},
			'contamination': {},
			'amplitude': {
				'wvf_extraction': {'ms_before': 1.0, 'ms_after': 1.0, 'max_spikes_per_unit': 500},
				'peak_sign': "both",
				'mode': "extremum",
				'filter': [100, 9_000]
			},
			'SNR': {
				'wvf_extraction': {'ms_before': 1.0, 'ms_after': 1.0, 'max_spikes_per_unit': 500},
				'peak_sign': "both",
				'mode': "extremum",
				'filter': [100, 9_000]
			},
			'sd_ratio': {
				'wvf_extraction': {'ms_before': 1.0, 'ms_after': 1.0, 'max_spikes_per_unit': 500},
				'spike_amplitudes_kwargs': {'peak_sign': "both"},
				'sd_ratio_kwargs': {},
				'filter': [100, 9_000]
			},
			'ISI_portion': {}
		}

		if attribute not in default_params:
			raise ValueError(f"Unknown attribute '{attribute}'.")
		params = default_params[attribute] | params

		recording = self.data.recording
		sorting = self.sorting
		if 'filter' in params:
			assert len(params['filter']) == 2, "The filter must be a list of 2 elements [min_cutoff, max_cutoff] (in Hz)."
			recording = spre.gaussian_filter(recording, *params['filter'], margin_sd=2)

		wvf_extractor = self.extract_waveforms(sub_folder=attribute, **params['wvf_extraction']) if 'wvf_extraction' in params \
						else si.WaveformExtractor(recording, sorting, allow_unfiltered=True)

		match attribute:
			case "firing_rate":  # Returns the firing rate of each unit (in Hz).
				n_spikes = {unit_id: len(sorting.get_unit_spike_train(unit_id)) for unit_id in sorting.unit_ids}
				firing_rates = {unit_id: n_spike * sorting.get_sampling_frequency() / recording.get_num_frames() for unit_id, n_spike in n_spikes.items()}
				return firing_rates

			case "contamination":  # Returns the estimated contamination of each unit.
				censored_period, refractory_period = params['refractory_period']
				contamination, _ = sqm.compute_refrac_period_violations(wvf_extractor, refractory_period_ms=refractory_period, censored_period_ms=censored_period)
				return contamination

			case "amplitude":  # Returns the amplitude of each unit on its best channel (unit depends on the wvf extractor 'return_scaled' parameter).
				params = utils.filter_kwargs(params, si.template_tools.get_template_extremum_amplitude)
				amplitudes = si.get_template_extremum_amplitude(wvf_extractor, **params)
				return amplitudes

			case "SNR":  # Returns the signal-to-noise ratio of each unit on its best channel.
				params = utils.filter_kwargs(params, sqm.compute_snrs)
				SNRs = sqm.compute_snrs(wvf_extractor, **params)
				return SNRs

			case "sd_ratio":  # Returns the standard deviation of the amplitude of spikes divided by the standard deviation on the same channel.
				_ = spost.compute_spike_amplitudes(wvf_extractor, **params['spike_amplitudes_kwargs'])
				sd_ratio = sqm.compute_sd_ratio(wvf_extractor, **params['sd_ratio_kwargs'])
				return sd_ratio

			case "ISI_portion":  # Returns the portion of consecutive spikes that are between a certain range (in ms).
				low, high = np.array(params['range']) * recording.sampling_frequency * 1e-3
				diff = {unit_id: np.diff(sorting.get_unit_spike_train(unit_id)) for unit_id in sorting.unit_ids}
				ISI_portion = {unit_id: np.nan if len(d) == 0 else (np.sum((low < d) & (d < high)) / len(d)) for unit_id, d in diff.items()}
				return ISI_portion

			case _:  # pragma: no cover (unreachable code)
				raise ValueError(f"Unknown attribute: '{attribute}'")

	def get_units_attribute_arr(self, attribute: str, params: dict) -> np.array:
		"""
		See MonoSortingModule.get_units_attribute.
		Returns the same value but as a numpy array rather than a dict.
		"""

		return np.array(list(self.get_units_attribute(attribute, params).values()))


@dataclass(slots=True)
class MultiSortingsModule(LussacModule):
	"""
	The abstract multi-sorting module class.
	This is for modules that work on multiple sortings at once.

	Attributes:
		name			Module's name (i.e. the key in the pipeline dictionary).
		data			Reference to Lussac data object.
		category		What category is used for the module (i.e. the key in the dictionary).
		export_sortings	Whether to export the sortings after the module is executed.
	"""

	data: MultiSortingsData

	@property
	def sortings(self) -> dict[str, si.BaseSorting]:
		"""
		Returns the sorting objects.

		@return sortings: dict[str, BaseSorting]
			The sorting objects.
		"""

		return self.data.sortings

	@property
	def logs_folder(self) -> pathlib.Path:
		"""
		Returns the logs directory for this module.

		@return logs_folder: Path
			Path to the logs directory.
		"""

		logs_folder = self.data.logs_folder / self.name / self.category
		logs_folder.mkdir(parents=True, exist_ok=True)

		return logs_folder

	@property
	def tmp_folder(self) -> pathlib.Path:
		"""
		Returns the temporary directory for this module.

		@return tmp_folder: Path
			Path to the temporary directory.
		"""

		tmp_folder = self.data.tmp_folder / self.name / self.category
		tmp_folder.mkdir(parents=True, exist_ok=True)

		return tmp_folder

	@abstractmethod
	def run(self, params: dict[str, Any]) -> dict[str, si.BaseSorting]:
		...

	def extract_waveforms(self, sub_folder: str | None = None, filter: list[float, float] | None = None, **params) -> si.WaveformExtractor:
		"""
		Aggregates all sortings and calls parent LussacModule.extract_waveforms.
		The returned WaveformExtractor has a variable 'renamed_unit_ids' which is a dict[str, dict[Any, Any]]
		where the first key is the analysis name, and the second one is the 'old' unit_id.
		"""

		aggregated_sortings = si.aggregate_units(list(self.sortings.values()))
		aggregated_sortings.annotate(name="aggregated_sortings")
		wvf_extractor = super(MultiSortingsModule, self).extract_waveforms(aggregated_sortings, sub_folder, filter, **params)

		wvf_extractor.sortings = self.sortings
		wvf_extractor.renamed_unit_ids = {}
		renamed_unit_id = 0
		for sorting_name in self.sortings.keys():
			wvf_extractor.renamed_unit_ids[sorting_name] = {}

			for unit_id in self.sortings[sorting_name].unit_ids:
				wvf_extractor.renamed_unit_ids[sorting_name][unit_id] = renamed_unit_id
				renamed_unit_id += 1

		return wvf_extractor



import copy
from dataclasses import dataclass
import glob
import logging
import os
import pathlib
import time
from typing import Type
import numpy as np
import numpy.typing as npt
from tqdm import tqdm
from lussac.core import LussacData, MonoSortingData, MultiSortingsData, MonoSortingModule, MultiSortingsModule, ModuleFactory
import spikeinterface.core as si


@dataclass(slots=True)
class LussacPipeline:
	"""
	The pipeline object for Lussac.

	Attributes:
		data			Reference to the data object.
		module_factory	Object to load and run all the modules.
	"""

	data: LussacData
	module_factory: ModuleFactory

	def __init__(self, data: LussacData) -> None:
		"""
		Creates a new LussacPipeline instance.

		@param data: LussacData
			The data object.
		"""

		self.data = data
		self.module_factory = ModuleFactory()

	def launch(self) -> None:
		"""
		Launches the Lussac's pipeline.
		"""

		for module_key, module_params in self.data.params['lussac']['pipeline'].items():
			logging.info('\n\n' + ('*'*34) + '\n')
			logging.info(f"{' ' + module_key + ' ':*^34}\n")
			logging.info(('*'*34) + '\n')

			if os.path.exists(f"{self.data.logs_folder}/{module_key}/sorting"):
				self.data.sortings = self._load_sortings(module_key)
				continue

			module_name = self._get_module_name(module_key)
			module = self.module_factory.get_module(module_name)

			if issubclass(module, MonoSortingModule):
				for category, params in module_params.items():
					logging.info(f"Running category '{category}':\n")
					self._run_mono_sorting_module(module, module_key, category, params)
			elif issubclass(module, MultiSortingsModule):
				self._run_multi_sortings_module(module, module_key, module_params)
			else:  # pragma: no cover (unreachable code)
				raise Exception("Error: Module does not inherit from MonoSortingModule or MultiSortingsModule.")

			self._save_sortings(module_key)

	def _run_mono_sorting_module(self, module: Type[MonoSortingModule], module_name: str, category: str, params: dict) -> None:
		"""
		Launches a mono-sorting module for a category on all sortings.

		@param module: MonoSortingModule
			The module class to use.
		@param module_name: str
			The module's name/key in the json file.
		@param category: str
			Run the module only on units from this category.
		@param params: dict
			The parameters for the module.
		"""

		for name, sorting in self.data.sortings.items():
			if 'sortings' in params and name not in params['sortings']:
				continue

			logging.info(f"\t- Sorting  {name:<18}")
			t1 = time.perf_counter()

			unit_ids = self.get_unit_ids_for_category(category, sorting)
			sub_sorting, other_sorting = self.split_sorting(sorting, unit_ids)

			data = MonoSortingData(self.data, sub_sorting)
			module_instance = module(module_name, data, category)
			params0 = copy.deepcopy(params)
			params0 = module_instance.update_params(params0)
			if 'sortings' in params0:
				del params0['sortings']

			sub_sorting = module_instance.run(params0)

			self.data.sortings[name] = self.merge_sortings(sub_sorting, other_sorting)

			t2 = time.perf_counter()
			logging.info(f" (Done in {t2-t1:.1f} s)\n")

	def _run_multi_sortings_module(self, module: Type[MultiSortingsModule], module_name: str, module_params: dict) -> None:
		"""
		Launches a multi-sorting module for a category.

		@param module: MultiSortingsModule
			The module class to use.
		@param module_name: str
			The module's name/key in the json file.
		@param module_params: dict
			The parameters for the module.
		"""

		new_sortings = {}
		for category, params in module_params.items():
			logging.info(f"Running category '{category}':\n")
			t1 = time.perf_counter()

			sub_sortings = {}
			for name, sorting in self.data.sortings.items():
				if 'sortings' in params and name not in params['sortings']:
					continue

				unit_ids = self.get_unit_ids_for_category(category, sorting)
				sub_sortings[name], _ = self.split_sorting(sorting, unit_ids)

			data = MultiSortingsData(self.data, sub_sortings)
			module_instance = module(module_name, data, category)
			params = copy.deepcopy(params)
			params = module_instance.update_params(params)

			sub_sortings = module_instance.run(params)

			for name, sub_sorting in sub_sortings.items():
				if name not in new_sortings:
					new_sortings[name] = sub_sorting
				else:
					new_sortings[name] = si.UnitsAggregationSorting([new_sortings[name], sub_sorting])

			t2 = time.perf_counter()
			logging.info(f"\tDone in {t2-t1:.1f} s\n")

		self.data.sortings = new_sortings

	def _save_sortings(self, module_name: str) -> None:
		"""
		Saves the current state of the sortings after a module run.

		@param module_name: str
			The module's name/key in the json file.
		"""

		for name, sorting in self.data.sortings.items():
			path = f"{self.data.logs_folder}/{module_name}/sorting/{name}.pkl"
			# sorting.dump_to_pickle(file_path=path, include_properties=True, relative_to=self.data.logs_folder)
			sorting.dump_to_pickle(file_path=path, include_properties=True)
			# TODO: Make relative paths work with pickle in SI.

	def _load_sortings(self, module_name: str) -> dict[str, si.BaseSorting]:
		"""
		Loads the sortings from a previous module run.

		@param module_name: str
			The module's name/key in the json file.
		@return sortings: dict[str, si.BaseSorting]
			The loaded sortings.
		"""

		logging.info("Loading sortings from previous run...\n")
		sortings_path = glob.glob(f"{self.data.logs_folder}/{module_name}/sorting/*.pkl")
		# sortings = {pathlib.Path(path).stem: si.load_extractor(path, base_folder=self.data.logs_folder) for path in tqdm(sortings_path)}
		sortings = {pathlib.Path(path).stem: si.load_extractor(path) for path in tqdm(sortings_path)}

		return sortings

	@staticmethod
	def _get_module_name(name: str) -> str:
		"""
		Gets the module name from the params dictionary key.
		Since all keys in a dict need to be different, this allows to use the same
		module multiple times by adding _{number} to the key.

		@param name: str
			Key from the params dictionary.
		@return module_name: str
			The module's name.
		"""

		if (split := name.split('_'))[-1].isnumeric():
			name = '_'.join(split[:-1])

		return name

	@staticmethod
	def get_unit_ids_for_category(category: str, sorting: si.BaseSorting) -> npt.NDArray[np.integer]:
		"""
		Gets all the unit ids for a given category.

		@param category: str
			The category to get the unit ids for.
		@param sorting: BaseSorting
			The sorting to get the unit ids from.
		@return unit_ids: np.ndarray[int]
			The unit ids for the given category.
		"""

		unit_ids = []
		units_category = sorting.get_property("lussac_category")

		categories = category.split('+')
		for cat in categories:
			if cat == "all":
				unit_ids.extend(sorting.unit_ids)
			elif cat == "rest":
				indices = np.where((units_category == '') | (units_category == None))[0]
				unit_ids.extend(sorting.unit_ids[indices])
			elif units_category is not None and len(units_category) > 0:
				indices = np.where(units_category == cat)[0]
				unit_ids.extend(sorting.unit_ids[indices])

		return np.sort(np.unique(unit_ids))

	@staticmethod
	def split_sorting(sorting: si.BaseSorting, unit_ids: npt.ArrayLike) -> tuple[si.BaseSorting, si.BaseSorting]:
		"""
		Splits a sorting into two based on the given unit ids.

		@param sorting: si.BaseSorting
			The sorting to split.
		@param unit_ids: ArrayLike[int]
			The unit ids of the first sorting.
		@return split_sortings: tuple[si.BaseSorting, si.BaseSorting]
			The split sortings.
		"""

		other_unit_ids = [unit_id for unit_id in sorting.get_unit_ids() if unit_id not in unit_ids]
		sorting1 = sorting.select_units(unit_ids)
		sorting2 = sorting.select_units(other_unit_ids)

		return sorting1, sorting2

	@staticmethod
	def merge_sortings(sorting1: si.BaseSorting, sorting2: si.BaseSorting) -> si.BaseSorting:
		"""
		Merges two split sortings into one.

		@param sorting1: si.BaseSorting
			The first sorting.
		@param sorting2: si.BaseSorting
			The second sorting.
		@return merged_sorting: si.BaseSorting
			The merged sorting.
		"""

		if sorting2.get_num_units() == 0:
			return sorting1
		if sorting1.get_num_units() == 0:
			return sorting2

		# Check for unit in sorting2 that has the same id as a unit in sorting1.
		rename_units = {}
		renamed = False
		for unit_id in sorting2.unit_ids:
			if unit_id in sorting1.unit_ids:
				rename_units[unit_id] = np.max([*sorting1.unit_ids, *sorting2.unit_ids, *rename_units.values()]) + 1
				renamed = True
			else:
				rename_units[unit_id] = unit_id

		if renamed:
			sorting2 = sorting2.select_units(list(rename_units.keys()), list(rename_units.values()))

		return si.UnitsAggregationSorting([sorting1, sorting2], renamed_unit_ids=[*sorting1.unit_ids, *sorting2.unit_ids])



import copy
import datetime
import logging
import os
import pathlib
import shutil
import sys
import tempfile
from dataclasses import dataclass
import numba
import numpy as np
import numpy.typing as npt
from plotly.offline.offline import get_plotlyjs
from lussac.utils import Utils
import probeinterface.io
import spikeinterface.core as si
import spikeinterface.curation as scur
import spikeinterface.extractors as se


class LussacData:
	"""
	The main data object for Lussac.

	Attributes:
		recording		A SpikeInterface recording object.
		sortings		A list of SpikeInterface sorting objects.
		params			A dictionary containing all of Lussac's parameters.
		_tmp_directory	The TemporaryDirectory object for Lussac.
	"""

	__slots__ = "recording", "sortings", "params", "_tmp_directory"
	recording: si.BaseRecording
	sortings: dict[str, si.BaseSorting]
	params: dict[str, dict]
	_tmp_directory: tempfile.TemporaryDirectory

	def __init__(self, recording: si.BaseRecording, sortings: dict[str, si.BaseSorting], params: dict[str, dict]) -> None:
		"""
		Creates a new LussacData instance.
		Loads all the necessary information from spike sorters output.

		@param recording: BaseRecording
			The recording object.
		@param sortings: dict[str, BaseSorting]
			A dict containing all the sorting objects (i.e. all the analyses).
			The keys are the analyses' names.
		@param params: dict
			The params.json file containing everything we need to know.
		"""

		self.recording = recording
		self.sortings = {name: scur.remove_excess_spikes(sorting.remove_empty_units(), recording) for name, sorting in sortings.items()}
		params['lussac']['pipeline'] = self._format_params(params['lussac']['pipeline'])
		self.params = params
		self._tmp_directory = self._setup_tmp_directory(params['lussac']['tmp_folder'])
		self._setup_logs_directory(params['lussac']['logs_folder'], params['lussac']['overwrite_logs'])

		if 'si_global_job_kwargs' in params['lussac']:
			si.set_global_job_kwargs(**params['lussac']['si_global_job_kwargs'])
			numba.set_num_threads(si.get_global_job_kwargs()['n_jobs'])

		targets = logging.StreamHandler(sys.stdout), logging.FileHandler(self.logs_folder / "lussac.logs")
		targets[0].terminator = ''
		targets[1].terminator = ''
		logging.basicConfig(format="%(message)s", level=logging.INFO, handlers=targets)
		logging.info(f"\nRunning Lussac!\n{datetime.datetime.today().strftime('%Y-%m-%d %H:%M:%S')}\n\n")

		self._sanity_check()

		Utils.sampling_frequency = recording.sampling_frequency
		Utils.t_max = recording.get_num_frames()

	def clone(self) -> 'LussacData':
		"""
		Returns a copy of the LussacData object.

		@return: LussacData
			A copy of the LussacData object.
		"""

		data = copy.copy(self)
		data.recording = data.recording.clone()
		data.sortings = {name: sorting.clone() for name, sorting in data.sortings.items()}
		data.params = copy.deepcopy(self.params)

		return data

	@property
	def tmp_folder(self) -> pathlib.Path:
		"""
		Returns the path to the temporary folder.
		This folder is deleted when Lussac exists (normally or because of a crash).

		@return tmp_folder: Path
			Path to the folder that store temporary files.
		"""

		return pathlib.Path(self._tmp_directory.name)

	@property
	def logs_folder(self) -> pathlib.Path:
		"""
		Returns the path to the logs folder.

		@return: logs_folder: Path
			Path to the folder that stores the logs.
		"""

		return pathlib.Path(self.params['lussac']['logs_folder'])

	@property
	def sampling_f(self) -> float:
		"""
		Returns the sampling frequency of the recording (in Hz).

		@return sampling_frequency: float
		"""

		return self.recording.sampling_frequency

	@property
	def num_sortings(self) -> int:
		"""
		Returns the number of sortings.

		@return num_sortings: int
			The number of sortings in the LussacData object.
		"""

		return len(self.sortings)

	def _sanity_check(self) -> None:
		"""
		Checks that everything seems correct in the recording and sortings.
		"""

		assert self.recording.get_num_frames() > 0
		assert self.recording.get_num_channels() > 0
		assert self.recording.get_probegroup() is not None

		for name, sorting in self.sortings.items():
			assert sorting.get_sampling_frequency() == self.sampling_f
			assert sorting.get_num_segments() == self.recording.get_num_segments()
			assert sorting.get_annotation("name") == name

			# Check that spike trains are valid.
			spike_vector = sorting.to_spike_vector()
			assert spike_vector['sample_index'][0] >= 0
			assert spike_vector['sample_index'][-1] < self.recording.get_num_frames()
			assert np.all(np.diff(spike_vector['sample_index']) >= 0)

	@staticmethod
	def _setup_probe(recording: si.BaseRecording, filename: str) -> si.BaseRecording:
		"""
		Loads the probe geometry into the 'recording' object.

		@param recording: BaseRecording
			Recording on which to load the probe geometry.
		@param filename: str
			Path to the JSON file containing the probe geometry (ProbeInterface format).
		@return probed_recording: BaseRecording
			The recording object with the probe geometry loaded.
		"""

		probe_group = probeinterface.io.read_probeinterface(filename)
		return recording.set_probegroup(probe_group)

	@staticmethod
	def _load_recording(params: dict) -> si.BaseRecording:
		"""
		Loads the recording from the given parameters.

		@param params: dict
			A dictionary containing Lussac's recording parameters.
		@return recording: BaseRecording
			The recording object.
		"""

		recording_extractor = se.extractorlist.get_recording_extractor_from_name(params['recording_extractor'])
		return recording_extractor(**params['extractor_params'])

	@staticmethod
	def _load_sortings(sortings_path: dict[str, str]) -> dict[str, si.BaseSorting]:
		"""
		Loads all the sortings (in Phy format) and return

		@param sortings_path: dict[str, str]
			Dict containing the name as key, and the path to all the
			phy folder / SpikeInterface extractor containing the spike sorted data as value.
		@return sortings: dict[str, BaseSorting]
			Dictionary containing the Phy sorting objects indexed by their name.
		"""

		sortings = {}
		for name, path in sortings_path.items():
			path = pathlib.Path(path)

			if not path.exists():
				raise FileNotFoundError(f"Could not find the sorting file {path}.")
			elif path.is_dir() and (path / "spike_times.npy").exists():
				sorting = se.PhySortingExtractor(path)
			else:
				sorting = si.load_extractor(path, base_folder=True)
				assert isinstance(sorting, si.BaseSorting)

			sorting.annotate(name=name)
			sortings[name] = sorting

		return sortings

	@staticmethod
	def _format_params(params: dict[str, dict]) -> dict[str, dict]:
		"""
		Formats the parameters' dictionary to take care of semicolons ';'.
		When a semicolon appears in the category, it duplicates the module and
		runs it separately for all categories separated by the semicolon.

		@param params: dict
			The parameters' dictionary for the Lussac pipeline.
		@return formatted_params: dict:
			The formatted parameters' dictionary.
		"""

		formatted_params = {}

		for module, value in params.items():
			if not isinstance(module, str):
				raise Exception(f"Error: Module {module} in params['lussac']['pipeline'] must be a string.")
			if not isinstance(value, dict):
				raise Exception(f"Error: params['lussac']['pipeline'][{module}] must map to a dict.")

			formatted_params[module] = {}

			for category, parameters in value.items():
				if not isinstance(category, str):
					raise Exception(f"Error: Category {category} in params['lussac']['pipeline'][{module}] must be a string.")
				if not isinstance(parameters, dict):
					raise Exception(f"Error: params['lussac']['pipeline'][{module}][{category}] must map to a dict.")

				categories = category.split(';')

				for cat in categories:
					formatted_params[module][cat] = copy.deepcopy(parameters)

		return formatted_params

	@staticmethod
	def _setup_tmp_directory(folder_path: str) -> tempfile.TemporaryDirectory:
		"""
		Created a directory for temporary files.
		This directory is deleted when Lussac exists (whether normally or because of a crash).

		@param folder_path: str
			Path for the temporary directory.
		@return temporary_directory: TemporaryDirectory
			The temporary directory object (from tempfile library).
		"""

		folder_path = pathlib.Path(folder_path)
		folder_path.parent.mkdir(parents=True, exist_ok=True)
		tmp_dir = tempfile.TemporaryDirectory(prefix=folder_path.name + '_', dir=folder_path.parent)

		return tmp_dir

	@staticmethod
	def _setup_logs_directory(folder_path: str, overwrite_logs: bool = False) -> None:
		"""
		Sets up the folder containing Lussac's logs.

		@param folder_path: str
			Path for the logs directory.
		@param overwrite_logs: bool
			If True, will erase the logs folder if it already exists.
		"""

		if overwrite_logs and os.path.exists(folder_path):
			shutil.rmtree(folder_path)

		os.makedirs(folder_path, exist_ok=True)

		# The javascript for plotly html files is about ~3 MB.
		# To not export it for each plot, store it in the logs and point to it in each html file.
		Utils.plotly_js_file = pathlib.Path(f"{folder_path}/plotly.min.js")
		if not os.path.exists(Utils.plotly_js_file):
			file = open(Utils.plotly_js_file, 'w+', encoding="utf-8")
			file.write(get_plotlyjs())
			file.close()

	@staticmethod
	def create_from_params(params: dict[str, dict]) -> 'LussacData':
		"""
		Creates a new LussacData object from the given parameters.

		@param params: dict
			Lussac's parameters.
		@return: LussacData
			The newly created LussacData object.
		"""

		recording = LussacData._load_recording(params['recording'])
		if 'probe_file' in params['recording']:
			recording = LussacData._setup_probe(recording, str(pathlib.Path(params['recording']['probe_file']).absolute()))
		sortings = LussacData._load_sortings(params['analyses'] if 'analyses' in params else {})

		return LussacData(recording, sortings, params)


@dataclass(slots=True)
class MonoSortingData:
	"""
	Allows easy manipulation of the LussacData object when working on only one sorting (or sub-sorting).

	Attributes:
		data	The main data object for Lussac.
		sorting	The sorting being used.
	"""

	data: LussacData
	sorting: si.BaseSorting

	@property
	def recording(self) -> si.BaseRecording:
		"""
		Returns the recording object.

		@return recording: BaseRecording
			The recording object.
		"""

		return self.data.recording

	@property
	def name(self) -> str:
		"""
		Returns the name of the current active sorting.

		@return name: str
			The name of the current active sorting.
		"""

		return self.sorting.get_annotation("name")

	@property
	def sampling_f(self) -> float:
		"""
		Returns the sampling frequency of the recording (in Hz).

		@return sampling_frequency: float
			The sampling frequency (in Hz).
		"""

		return self.data.sampling_f

	@property
	def tmp_folder(self) -> pathlib.Path:
		"""
		Returns the path to the temporary folder.
		This folder is deleted when Lussac exists (normally or because of a crash).

		@return tmp_folder: Path
			Path to the folder that store temporary files.
		"""

		return self.data.tmp_folder

	@property
	def logs_folder(self) -> pathlib.Path:
		"""
		Returns the path to the logs folder.

		@return logs_folder: Path
			Path to the folder that stores Lussac's logs.
		"""

		return self.data.logs_folder

	def get_unit_spike_train(self, unit_id: int) -> npt.NDArray[np.integer]:
		"""
		Returns the spike_train (i.e. an array containing all the spike timings)
		of a given unit.

		@param unit_id: int
			The cluster's ID of which to return the spike train.
		@return spike_train: np.ndarray
		"""

		return self.sorting.get_unit_spike_train(unit_id)


@dataclass(slots=True)
class MultiSortingsData:
	"""
	Allows easy manipulation of the LussacData object when working on multiple sortings (or sub-sortings).

	Attributes:
		data		The main data object for Lussac.
		sortings	The sortings being used.
	"""

	data: LussacData
	sortings: dict[str, si.BaseSorting]

	@property
	def recording(self) -> si.BaseRecording:
		"""
		Returns the recording object.

		@return recording: BaseRecording
			The recording object.
		"""

		return self.data.recording

	@property
	def num_sortings(self) -> int:
		"""
		Returns the number of sortings.

		@return num_sortings: int
			The number of sortings.
		"""

		return len(self.sortings)

	@property
	def tmp_folder(self) -> pathlib.Path:
		"""
		Returns the path to the temporary folder.
		This folder is deleted when Lussac exists (normally or because of a crash).

		@return tmp_folder: Path
			Path to the folder that store temporary files.
		"""

		return self.data.tmp_folder

	@property
	def logs_folder(self) -> pathlib.Path:
		"""
		Returns the path to the logs folder.

		@return logs_folder: Path
			Path to the folder that stores Lussac's logs.
		"""

		return self.data.logs_folder



from typing import Any
import networkx as nx
import numpy as np
from overrides import override
import plotly.graph_objects as go
from lussac.core import MonoSortingModule
import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.curation as scur
import spikeinterface.qualitymetrics as sqm
from spikeinterface.curation.auto_merge import normalize_correlogram


class MergeUnits(MonoSortingModule):
	"""
	Detects and merges units that come from the same neuron if there is an increase in score.
	If there is a decrease in score, the worse unit is removed.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'refractory_period': [0.2, 1.0],
			'wvf_extraction': {
				'ms_before': 1.0,
				'ms_after': 1.5,
				'max_spikes_per_unit': 2_000,
				'filter': [100, 9000]
			},
			'auto_merge_params': {
				'bin_ms': 0.05,
				'window_ms': 150,
				'corr_diff_thresh': 0.16,
				'template_diff_thresh': 0.25,
				'firing_contamination_balance': 2.5
			}
		}

	@override
	def update_params(self, params: dict[str, Any]) -> dict[str, Any]:
		params = super().update_params(params)

		params['auto_merge_params']['censored_period_ms'] = params['refractory_period'][0]
		params['auto_merge_params']['refractory_period_ms'] = params['refractory_period'][1]

		return params

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		wvf_extractor = self.extract_waveforms(sparse=False, **params['wvf_extraction'])
		potential_merges, extra_outputs = scur.get_potential_auto_merge(wvf_extractor, extra_outputs=True, **params['auto_merge_params'])

		sorting = self._remove_splits(self.sorting, extra_outputs, params)
		sorting = self._merge(sorting, potential_merges, params)
		self.plot_merging(potential_merges, wvf_extractor, extra_outputs, params['auto_merge_params'])

		return sorting

	def _remove_splits(self, sorting: si.BaseSorting, extra_outputs: dict, params: dict[str, Any]) -> si.BaseSorting:
		"""
		Remove units that are split but decrease the score if merged.
		When such a pair is detected, remove the unit with the lowest score.
		This unit usually is a contaminated unit.
		TODO: Create a plot of removed units (logs).

		@param sorting: si.BaseSorting
			The sorting object.
		@param extra_outputs: dict
			Extra outputs given by the merging process.
		@param params: dict
			Parameters of the merging process.
			i.e. params['auto_merge_params']
		@return: si.BaseSorting
			Sorting with the units removed.
		"""
		t_c, t_r = params['refractory_period']
		k = params['auto_merge_params']['firing_contamination_balance']

		units_to_remove = []
		wvf_extractor = si.WaveformExtractor(self.recording, sorting, allow_unfiltered=True)
		contamination, _ = sqm.compute_refrac_period_violations(wvf_extractor, refractory_period_ms=t_r, censored_period_ms=t_c)

		for pair in extra_outputs['pairs_decreased_score']:
			unit1, unit2 = pair
			score_unit1 = len(sorting.get_unit_spike_train(unit1)) * (1 - (k+1) * contamination[unit1])
			score_unit2 = len(sorting.get_unit_spike_train(unit2)) * (1 - (k+1) * contamination[unit2])
			worse_unit = unit1 if score_unit1 < score_unit2 else unit2
			units_to_remove.append(worse_unit)  # TODO: Only remove if unconnected in 'potential_merges' graph?

		return sorting.select_units([unit_id for unit_id in sorting.unit_ids if unit_id not in units_to_remove])

	def _merge(self, sorting: si.BaseSorting, potential_merges: list[tuple], params: dict[str, Any]) -> si.BaseSorting:
		"""
		Merges units based on a list of potential merges.
		Creates a graph with all the units, where potential merges are the edges.
		For each connected component (i.e. each putative neuron), go through all the edges and compute the
		score of the merge. If the score increases (over lone unit scores), merges the best edge.
		This is done iteratively until either one unit remains or no increase is detected.
		The unit not merged (because of decrease in score) are deleted.

		@param sorting: si.BaseSorting
			The sorting object.
		@param potential_merges: list[tuple]
			List of potential merges.
		@param params: dict
			Parameters of the merging process.
			i.e. params['auto_merge_params']
		@return: si.BaseSorting
			Sorting with the splits merged and removed.
		"""
		t_c, t_r = params['refractory_period']
		k = params['auto_merge_params']['firing_contamination_balance']

		wvf_extractor = si.WaveformExtractor(self.recording, sorting, allow_unfiltered=True)
		contamination, _ = sqm.compute_refrac_period_violations(wvf_extractor, refractory_period_ms=t_r, censored_period_ms=t_c)
		sorting = scur.CurationSorting(sorting, properties_policy="keep")

		graph = nx.Graph()
		for potential_merge in potential_merges:
			unit1, unit2 = potential_merge
			if unit1 not in sorting.sorting.unit_ids or unit2 not in sorting.sorting.unit_ids:
				continue  # Can happen if a unit is removed in '_remove_splits'.

			for unit in [unit1, unit2]:
				if unit not in graph:
					score = len(sorting.sorting.get_unit_spike_train(unit)) * (1 - (k+1) * contamination[unit])
					graph.add_node(unit, score=score)

			graph.add_edge(*potential_merge)

		# For each putative neuron, merge the units.
		for units in nx.connected_components(graph):
			subgraph = graph.subgraph(units).copy()

			while len(subgraph) > 1:
				highest_score = max(dict(subgraph.nodes(data="score")).values())
				merge = None

				for unit1, unit2 in subgraph.edges:
					sorting_merged = scur.MergeUnitsSorting(sorting.sorting, [[unit1, unit2]], new_unit_ids=[unit1], delta_time_ms=t_c).select_units([unit1])
					wvf_extractor = si.WaveformExtractor(self.recording, sorting_merged, allow_unfiltered=True)
					C = sqm.compute_refrac_period_violations(wvf_extractor, refractory_period_ms=t_r, censored_period_ms=t_c)[0][unit1]
					score = len(sorting_merged.get_unit_spike_train(unit1)) * (1 - (k+1) * C)

					if score > highest_score:
						highest_score = score
						merge = (unit1, unit2)

				if merge is None:
					scores = dict(subgraph.nodes(data="score"))
					best_unit = max(scores, key=scores.get)
					sorting.remove_units([x for x in subgraph.nodes if x != best_unit])
					break

				unit1, unit2 = merge
				sorting.merge(merge, new_unit_id=unit1)
				subgraph = nx.contracted_nodes(subgraph, unit1, unit2, self_loops=False)
				subgraph.nodes[unit1]['score'] = highest_score

		return sorting.sorting

	def plot_merging(self, potential_merges: list[tuple], wvf_extractor: si.WaveformExtractor, extra_outputs: dict[str, Any], params: dict[str, Any]) -> None:
		"""
		Makes different plots about the merging process.

		@param potential_merges: list[tuple]
			List of potential merges pairwise.
		@param wvf_extractor: WaveformExtractor
			Waveform extractor used to extract the waveforms.
		@param extra_outputs: dict[str, Any]
			Extra outputs given by the merging process.
		@param params: dict[str, Any]
			Parameters given to the merging process.
			i.e. params['auto_merge_params']
		"""

		self.plot_results(potential_merges, extra_outputs, params, wvf_extractor)
		self.plot_difference_matrix(extra_outputs, params)

	def plot_results(self, potential_merges: list[tuple], extra_outputs: dict[str, Any], params: dict[str, Any], wvf_extractor: si.WaveformExtractor) -> None:
		"""
		Plots the result of the merging process (i.e. the pairs merged or closed to be merged).

		@param potential_merges: list[tuple]
			List of potential merges pairwise (containing the unit ids).
		@param extra_outputs: dict[str, Any]
			The extra outputs given by the merging process.
		@param params: dict[str, Any]
			The parameters given to the merging process.
		@param wvf_extractor: WaveformExtractor
			The waveform extractor used by the merging process.
		"""
		bins = extra_outputs['bins']
		correlograms = extra_outputs['correlograms']
		correlograms_smoothed = extra_outputs['correlograms_smoothed']
		correlogram_diff = extra_outputs['correlogram_diff']
		templates_diff = extra_outputs['templates_diff']
		window_sizes = extra_outputs['win_sizes']
		corr_diff_threshold = params['corr_diff_thresh']

		fig = go.Figure().set_subplots(rows=2, cols=2)
		bins = bins[:-1] + (bins[1] - bins[0]) / 2
		t_axis = np.arange(-wvf_extractor.nbefore, wvf_extractor.nafter) / wvf_extractor.sampling_frequency * 1e3
		wvfs_unit = "µV" if wvf_extractor.return_scaled else "A.U."
		labels = []
		args = []

		for k in range(correlograms.shape[0]**2):
			i, j = k // correlograms.shape[0], k % correlograms.shape[0]
			if i == j:
				continue
			if correlogram_diff[i, j] > corr_diff_threshold + 0.05 or np.isnan(correlogram_diff[i, j]):
				continue

			unit_id_1 = self.sorting.unit_ids[i]
			unit_id_2 = self.sorting.unit_ids[j]
			color = "black" if (unit_id_1, unit_id_2) in potential_merges or (unit_id_2, unit_id_1) in potential_merges else "red"

			annotation_gt = {}
			if 'gt_label' in wvf_extractor.sorting.get_property_keys():
				gt_label_1 = wvf_extractor.sorting.get_unit_property(unit_id_1, 'gt_label')
				gt_label_2 = wvf_extractor.sorting.get_unit_property(unit_id_2, 'gt_label')
				annotation_gt = {
					'x': 1.0,
					'y': 1.05,
					'xref': "paper",
					'yref': 'paper',
					'xanchor': 'right',
					'yanchor': 'top',
					'text': f"GT {unit_id_1}: {gt_label_1}<br />GT {unit_id_2}: {gt_label_2}",
					'font': {
						'color': "rgba(73, 42, 189, 1.0)"
					},
					'showarrow': False
				}

			labels.append(f"Units {unit_id_1} & {unit_id_2}")
			args.append({'title.text': f"Units {unit_id_1} & {unit_id_2}: corr_diff = {correlogram_diff[i, j]:.1%} ; temp_diff = {templates_diff[i, j]:.1%}",
						 'title.font.color': color,
						 'annotations': [annotation_gt]})

			fig.add_trace(go.Scatter(
				x=bins,
				y=correlograms[i, i, :],
				mode="lines",
				name=f"Auto-corr unit {unit_id_1}",
				marker_color="CornflowerBlue"
			), row=1, col=1)
			fig.add_trace(go.Scatter(
				x=bins,
				y=correlograms[j, j, :],
				mode="lines",
				name=f"Auto-corr unit {unit_id_2}",
				marker_color="LightSeaGreen"
			), row=1, col=1)
			fig.add_trace(go.Scatter(
				x=bins,
				y=correlograms[i, j, :],
				mode="lines",
				name=f"Cross-corr unit {unit_id_1} - {unit_id_2}",
				marker_color="Crimson"
			), row=1, col=1)

			fig.add_trace(go.Scatter(
				x=bins,
				y=normalize_correlogram(correlograms_smoothed[i, i, :]),
				mode="lines",
				name=f"Auto-corr unit {unit_id_1}",
				marker_color="CornflowerBlue"
			), row=1, col=2)
			fig.add_trace(go.Scatter(
				x=bins,
				y=normalize_correlogram(correlograms_smoothed[j, j, :]),
				mode="lines",
				name=f"Auto-corr unit {unit_id_2}",
				marker_color="LightSeaGreen"
			), row=1, col=2)
			fig.add_trace(go.Scatter(
				x=bins,
				y=normalize_correlogram(correlograms_smoothed[i, j, :]),
				mode="lines",
				name=f"Cross-corr unit {unit_id_1} - {unit_id_2}",
				marker_color="Crimson"
			), row=1, col=2)
			# TODO: Plot window size for correlogram.

			templates1 = wvf_extractor.get_template(unit_id_1)
			templates2 = wvf_extractor.get_template(unit_id_2)
			best_channels = np.argsort(np.max(np.abs(templates1) + np.abs(templates2), axis=0))[::-1]
			for i in range(2):
				channel_id = wvf_extractor.recording.get_channel_ids()[best_channels[i]]
				fig.add_trace(go.Scatter(
					x=t_axis,
					y=templates1[:, best_channels[i]],
					mode="lines",
					name=f"Template unit {unit_id_1} (channel {channel_id})",
					marker_color="CornflowerBlue"
				), row=2, col=1+i)
				fig.add_trace(go.Scatter(
					x=t_axis,
					y=templates2[:, best_channels[i]],
					mode="lines",
					name=f"Template unit {unit_id_2} (channel {channel_id})",
					marker_color="LightSeaGreen"
				), row=2, col=1+i)

		fig.update_xaxes(title_text="Time (ms)", row=1, col=1)
		fig.update_xaxes(title_text="Time (ms)", matches='x', row=1, col=2)
		fig.update_xaxes(title_text="Time (ms)", row=2, col=1)
		fig.update_xaxes(title_text="Time (ms)", matches='x3', row=2, col=2)
		fig.update_yaxes(title_text=f"Voltage ({wvfs_unit})", row=2, col=1)
		fig.update_yaxes(title_text=f"Voltage ({wvfs_unit})", matches='y3', row=2, col=2)

		utils.plotting.plot_sliders(fig, 10, labels, f"{self.logs_folder}/results", args=args)

	def plot_difference_matrix(self, extra_outputs: dict[str, Any], params: dict[str, Any]) -> None:
		"""
		Plots the difference matrix between units pairwise.
		i.e. for each pair, plots the correlogram (x-axis) and template (y-axis) difference.

		@param extra_outputs: dict[str, Any]
			Extra outputs given by the merging process.
		@param params: dict[str, Any]
			Parameters given to the merging process.
			i.e. params['auto_merge_params']
		"""
		correlogram_diff = extra_outputs['correlogram_diff']
		templates_diff = extra_outputs['templates_diff']
		np.save(f"{self.logs_folder}/difference_matrix.npy", np.dstack((correlogram_diff, templates_diff)))

		unit_ids = self.sorting.unit_ids
		N = len(unit_ids)
		text = [f"Units {unit_ids[i//N]} - {unit_ids[i%N]}" for i in range(len(correlogram_diff)**2)]

		fig = go.Figure()

		fig.add_trace(go.Scatter(
			x=correlogram_diff.flatten(),
			y=templates_diff.flatten(),
			mode="markers",
			text=text,
			marker_color="CornflowerBlue"
		))

		fig.add_shape(
			type="rect",
			x0=0, x1=params['corr_diff_thresh'],
			y0=0, y1=params['template_diff_thresh'],
			line={'color': "Crimson", 'dash': "dash"}
		)

		fig.update_xaxes(title_text="Correlograms difference")
		fig.update_yaxes(title_text="Templates difference", range=[-0.03, 1.03])

		utils.plotting.export_figure(fig, f"{self.logs_folder}/difference_matrix.html")



import itertools
import math
import pickle
from typing import Any
from overrides import override
import networkx as nx
import numpy as np
from lussac.core import MultiSortingsModule
import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.curation as scur
from spikeinterface.curation.curation_tools import find_duplicated_spikes
import spikeinterface.postprocessing as spost
import spikeinterface.qualitymetrics as sqm


class MergeSortings(MultiSortingsModule):
	"""
	Merges the sortings into a single one.
	"""

	aggregated_wvf_extractor: si.WaveformExtractor  # Waveform extractor for all analyses.

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'refractory_period': [0.2, 1.0],
			'max_shift': 1.33,
			'require_multiple_sortings_match': True,
			'max_units_merge': 1,
			'similarity': {
				'min_similarity': 0.3,
				'window': 0.2
			},
			'correlogram_validation': {
				'max_time': 70.0,
				'gaussian_std': 0.6,
				'gaussian_truncate': 5.0
			},
			'waveform_validation': {
				'wvf_extraction': {
					'ms_before': 1.0,
					'ms_after': 2.0,
					'max_spikes_per_unit': 1_000,
					'filter': [250.0, 6_000.0]
				},
				'num_channels': 5
			},
			'merge_check': {
				'cross_cont_threshold': 0.10
			},
			'clean_edges': {
				'template_diff_threshold': 0.10,
				'corr_diff_threshold': 0.12,
				'cross_cont_threshold': 0.06
			}
		}

	@override
	def update_params(self, params: dict[str, Any]) -> dict[str, Any]:
		params = super().update_params(params)

		params['similarity']['window'] = int(round(params['similarity']['window'] * 1e-3 * self.sampling_f))
		params['correlogram_validation']['max_time'] = int(round(params['correlogram_validation']['max_time'] * 1e-3 * self.sampling_f))
		params['correlogram_validation']['gaussian_std'] = params['correlogram_validation']['gaussian_std'] * 1e-3 * self.sampling_f
		params['correlogram_validation']['censored_period'] = params['similarity']['window']
		params['waveform_validation']['wvf_extraction']['ms_before'] += params['max_shift']
		params['waveform_validation']['wvf_extraction']['ms_after']  += params['max_shift']
		params['max_shift'] = int(round(params['max_shift'] * 1e-3 * self.sampling_f))

		return params

	@override
	def run(self, params: dict[str, Any]) -> dict[str, si.BaseSorting]:
		self.data.sortings = {name: sorting.remove_empty_units() for name, sorting in self.sortings.items()}

		self.aggregated_wvf_extractor = self.extract_waveforms(sparse=False, **params['waveform_validation']['wvf_extraction'])
		cross_shifts = self.compute_cross_shifts(params['max_shift'])

		similarity_matrices = self._compute_similarity_matrices(cross_shifts, params)
		graph = self._compute_graph(similarity_matrices, params)
		self.compute_correlogram_difference(graph, cross_shifts, params['correlogram_validation'])
		self.compute_waveform_difference(graph, cross_shifts, params)

		if params['merge_check']:
			self.remove_merged_units(graph, cross_shifts, params['refractory_period'], params['merge_check'])
		self.clean_edges(graph, cross_shifts, params)
		if len(self.sortings) >= 4:
			self.separate_communities(graph)
		self._save_graph(graph, "final_graph")

		merged_sorting = self.merge_sortings(graph, params)

		return {'merged_sorting': merged_sorting}

	def compute_cross_shifts(self, max_shift: int) -> dict[str, dict[str, np.ndarray]]:
		"""
		Computes the cross shifts between units of different sortings.
		This is important when comparing units to make sure they are aligned properly between them.
		If two units don't come from the same neuron, the computed cross shift should be 0.
		A cross shift of -2 means the first unit is generally 2 samples earlier than the second unit.

		@param max_shift: int
			The maximum shift to consider (in samples).
		@return: dict[str, dict[str, np.ndarray]]
			For each pair of analyses, the cross-shifts matrix.
		"""

		spike_vectors = {name: sorting.to_spike_vector() for name, sorting in self.sortings.items()}
		cross_shifts = {name1: {} for name1 in self.sortings.keys()}

		for i, (name1, sorting1) in enumerate(self.sortings.items()):
			for j, (name2, sorting2) in enumerate(self.sortings.items()):
				if i >= j:
					continue

				shifts = utils.compute_cross_shift_from_vector(spike_vectors[name1], spike_vectors[name2], max_shift)

				cross_shifts[name1][name2] = shifts
				cross_shifts[name2][name1] = -shifts.T

		return cross_shifts

	def _compute_similarity_matrices(self, cross_shifts: dict[str, dict[str, np.ndarray | None]], params: dict[str, Any]) -> dict[str, dict[str, np.ndarray]]:
		"""
		Computes the similarity matrix between all sortings.

		@param cross_shifts: dict[str, dict[str, np.ndarray | None]]
			The cross-shifts between units.
			If None, will default to no shift.
		@param params: dict
			The parameters of the merge_sorting module.
		@return similarity_matrices: dict[str, dict[str, np.ndarray]]
			The similarity matrices [sorting1, sorting2, similarity_matrix].
		"""
		window = params['similarity']['window']
		censored_period = params['refractory_period'][0]

		similarity_matrices = {}
		spike_vectors = {name: scur.remove_duplicated_spikes(sorting, censored_period).to_spike_vector() for name, sorting in self.sortings.items()}
		n_spikes = {name: np.array(list(sorting.count_num_spikes_per_unit().values())) for name, sorting in self.sortings.items()}

		for name1, sorting1 in self.sortings.items():
			similarity_matrices[name1] = {}
			for name2, sorting2 in self.sortings.items():
				if name1 == name2 or name2 in similarity_matrices[name1]:
					continue
				if name2 not in similarity_matrices:
					similarity_matrices[name2] = {}

				coincidence_matrix = utils.compute_coincidence_matrix_from_vector(spike_vectors[name1], spike_vectors[name2], window, cross_shifts[name1][name2])

				similarity_matrix = utils.compute_similarity_matrix(coincidence_matrix, n_spikes[name1], n_spikes[name2], window)
				similarity_matrices[name1][name2] = similarity_matrix
				similarity_matrices[name2][name1] = similarity_matrix.T

		return similarity_matrices

	def _compute_graph(self, similarity_matrices: dict[str, dict[str, np.ndarray]], params: dict) -> nx.Graph:
		"""
		Creates a graph containing all the units from all the sortings,
		and edges between units that are similar.

		@param similarity_matrices: dict[str, dict[str, np.ndarray]]
			The similarity matrices [sorting1, sorting2, similarity_matrix].
		@param params: dict
			The parameters of the merge_sorting module.
		@return graph: nx.Graph
			The graph containing the edged for each pair of similar units.
		"""

		censored_period, refractory_period = params['refractory_period']

		# Populating the graph with all the nodes (i.e. all the units) with properties.
		graph = nx.Graph()
		spost.compute_spike_amplitudes(self.aggregated_wvf_extractor, peak_sign="both", return_scaled=self.recording.has_scaled())
		contamination, _ = sqm.compute_refrac_period_violations(self.aggregated_wvf_extractor, refractory_period_ms=refractory_period, censored_period_ms=censored_period)
		sd_ratio = sqm.compute_sd_ratio(self.aggregated_wvf_extractor)
		snrs = sqm.compute_snrs(self.aggregated_wvf_extractor, peak_sign="both", peak_mode="extremum")

		for (name, sorting) in self.sortings.items():
			for unit_id in sorting.unit_ids:
				new_unit_id = self.aggregated_wvf_extractor.renamed_unit_ids[name][unit_id]

				attr = {key: sorting.get_unit_property(unit_id, key) for key in sorting.get_property_keys() if key.startswith('gt_')}
				attr['contamination'] = contamination[new_unit_id]
				attr['sd_ratio'] = sd_ratio[new_unit_id]
				attr['SNR'] = snrs[new_unit_id]
				graph.add_node((name, unit_id), **attr)

		# Connecting nodes using the similarity.
		for i, (name1, sorting1) in enumerate(self.sortings.items()):
			for j, (name2, sorting2) in enumerate(self.sortings.items()):
				if j <= i:
					continue

				for unit_ind1, unit_id1 in enumerate(sorting1.unit_ids):
					for unit_ind2, unit_id2 in enumerate(sorting2.unit_ids):
						if (similarity := similarity_matrices[name1][name2][unit_ind1, unit_ind2]) >= params['similarity']['min_similarity']:
							graph.add_edge((name1, unit_id1), (name2, unit_id2), similarity=similarity)

		if params['require_multiple_sortings_match']:
			for node in dict(graph.nodes):
				if len(graph.edges(node)) == 0:
					graph.remove_node(node)

		self._save_graph(graph, "similarity_graph")
		return graph

	def _save_graph(self, graph: nx.Graph, name: str) -> None:
		"""
		Saves the current state of the graph in the pickle format.

		@param graph: nx.Graph
			The graph to save.
		@param name: str
			The name of the file (without the .pkl extension)
		"""

		with open(f"{self.logs_folder}/{name}.pkl", 'wb+') as file:
			pickle.dump(graph, file, protocol=pickle.HIGHEST_PROTOCOL)

	def remove_merged_units(self, graph: nx.Graph, cross_shifts: dict[str, dict[str, np.ndarray]], refractory_period, params: dict[str, Any]) -> None:
		"""
		Detects and remove merged units from the graph.
		For each connected components (i.e. connected sub-graph communities), look at each node. If a node is connected
		to two nodes coming from the same sorting, then either the original node is a merged unit, or the two nodes are
		split units. In the first case, (under the hypothesis that the 2 neurons are not correlated), the cross-contamination
		between the two nodes should be high (around 100%), whereas in the second case, the cross contamination should be
		low (around 0%). If the cross-contamination crosses a threshold, the original node is labeled as a merged unit
		and all of its edges are removed.

		@param graph: nx.Graph
			The graph containing all the units and connected by their similarity.
		@param cross_shifts: dict[str, dict[str, np.ndarray]]
			The cross-shifts between units.
		@param refractory_period: float
			The (censored_period, refractory_period) in ms.
		@param params: dict[str, Any]
			The parameters for the remove_merge function.
		"""

		logs = open(f"{self.logs_folder}/merged_units_logs.txt", 'w+')
		nodes_to_remove = []

		for node in graph.nodes:
			for node1, node2 in itertools.combinations(graph.neighbors(node), 2):
				sorting_name, unit_id = node
				sorting1_name, unit_id1 = node1
				sorting2_name, unit_id2 = node2

				unit_ind = self.sortings[sorting_name].id_to_index(unit_id)
				unit_ind1 = self.sortings[sorting1_name].id_to_index(unit_id1)
				unit_ind2 = self.sortings[sorting2_name].id_to_index(unit_id2)

				if sorting1_name != sorting2_name:
					continue

				spike_train = self.sortings[sorting_name].get_unit_spike_train(unit_id)
				spike_train1 = self.sortings[sorting1_name].get_unit_spike_train(unit_id1) + cross_shifts[sorting_name][sorting1_name][unit_ind, unit_ind1]
				spike_train2 = self.sortings[sorting2_name].get_unit_spike_train(unit_id2) + cross_shifts[sorting_name][sorting2_name][unit_ind, unit_ind2]
				C = graph.nodes[node]['contamination']
				C1 = graph.nodes[node1]['contamination']
				C2 = graph.nodes[node2]['contamination']
				sd = graph.nodes[node]['sd_ratio']
				sd1 = graph.nodes[node1]['sd_ratio']
				sd2 = graph.nodes[node2]['sd_ratio']
				if C2 < C1:
					spike_train1, spike_train2 = spike_train2, spike_train1
				cross_cont, p_value = utils.estimate_cross_contamination(spike_train1, spike_train2, refractory_period, limit=params['cross_cont_threshold'])

				logs.write(f"\nUnit {node} is connected to {node1} and {node2}:\n")
				logs.write(f"\tcross-cont = {cross_cont:.2%} (p_value={p_value:.2e})\n")
				logs.write(f"\tC = {C:.1%} ; C1 = {C1:.1%} ; C2 = {C2:.1%}\n")
				logs.write(f"\tsd = {sd:.3f} ; sd1 = {sd1:.3f} ; sd2 = {sd2:.3f}\n")
				if p_value > 5e-3:  # No problem detected, node1 and node2 are probably just a split.
					continue

				cases_removed = np.array([  # If at least one is true, 'node' is removed
					p_value < 1e-8 and C > max(C1, C2),
					p_value < 5e-3 and sd > max(sd1, sd2) and C > max(C1, C2)
				], dtype=bool)
				if np.any(cases_removed):
					nodes_to_remove.append(node)
					break  # Don't need to check this node again.
				else:  # Couldn't definitely say that 'node' is a merged unit. Check whether 'node1' or 'node2' is bad.
					cc1, p1 = utils.estimate_cross_contamination(spike_train, spike_train1, (0.4, 1.8), limit=0.10)
					cc2, p2 = utils.estimate_cross_contamination(spike_train, spike_train2, (0.4, 1.8), limit=0.10)
		
					if p1 < 5e-3 and cc2 < 0.1 and C1 > C + 0.03:
						nodes_to_remove.append(node1)
					elif p2 < 5e-3 and cc1 < 0.1 and C2 > C + 0.03:
						nodes_to_remove.append(node2)

		if len(nodes_to_remove) > 0:
			logs.write("\nRemoved units:\n")
			for node in set(nodes_to_remove):
				graph.remove_node(node)

				sorting_name, unit_id = node
				label = f" -- {self.sortings[sorting_name].get_unit_property(unit_id, 'gt_label')}" if 'gt_label' in self.sortings[sorting_name].get_property_keys() else ''
				logs.write(f"\t- {node}{label}\n")

		logs.close()

	def compute_correlogram_difference(self, graph: nx.Graph, cross_shifts: dict[str, dict[str, np.ndarray]], params: dict[str, Any]) -> None:
		"""
		Computes the correlogram difference for each edge of the graph, and adds it as an attribute to the edge.

		@param graph: nx.Graph
			The graph containing all the units and connected by their similarity.
		@param cross_shifts: dict[str, dict[str, np.ndarray]]
			The cross-shifts between units.
		@param params: dict
			The parameters for the correlogram difference.
		"""

		N = math.ceil(params['gaussian_std'] * params['gaussian_truncate'])
		gaussian = utils.gaussian_pdf(np.arange(-N, N+1), 0.0, params['gaussian_std'])

		for nodes in nx.connected_components(graph):  # For each community.
			nodes = list(nodes)
			if len(nodes) <= 1:
				continue

			subgraph = graph.subgraph(nodes)

			auto_correlograms = {}
			for node in nodes:
				sorting_name, unit_id = node
				spike_train = self.sortings[sorting_name].get_unit_spike_train(unit_id)
				auto_corr = spost.compute_autocorrelogram_from_spiketrain(spike_train, window_size=params['max_time'], bin_size=1)
				auto_correlograms[node] = np.convolve(auto_corr, gaussian, mode="same")

			for node1, node2 in subgraph.edges:
				sorting1_name, unit_id1 = node1
				sorting2_name, unit_id2 = node2
				unit_ind1 = self.sortings[sorting1_name].id_to_index(unit_id1)
				unit_ind2 = self.sortings[sorting2_name].id_to_index(unit_id2)

				shift = cross_shifts[sorting1_name][sorting2_name][unit_ind1, unit_ind2]
				spike_train1 = self.sortings[sorting1_name].get_unit_spike_train(unit_id1)
				spike_train2 = self.sortings[sorting2_name].get_unit_spike_train(unit_id2) + shift
				cross_corr = spost.compute_crosscorrelogram_from_spiketrain(spike_train1, spike_train2, window_size=params['max_time'], bin_size=1)

				middle = len(cross_corr) // 2
				cross_corr[middle-params['censored_period']:middle+params['censored_period']] = 0  # Remove duplicates before filtering.
				cross_corr = np.convolve(cross_corr, gaussian, mode="same")

				corr_diff = utils.compute_correlogram_difference(auto_correlograms[node1], auto_correlograms[node2], cross_corr, len(spike_train1), len(spike_train2))
				graph[node1][node2]['corr_diff'] = corr_diff

	def compute_waveform_difference(self, graph: nx.Graph, cross_shifts: dict[str, dict[str, np.ndarray]], params: dict[str, Any]) -> None:
		"""
		Computes the waveform difference for each edge of the graph, and adds it as an attribute to the edge.

		@param graph: nx.Graph
			The graph containing all the units and connected by their similarity.
		@param cross_shifts: dict[str, dict[str, np.ndarray]]
			The cross-shifts between units.
		@param params: dict
			The parameters for the remove_merge function.
		"""

		n_channels = params['waveform_validation']['num_channels']
		margin = params['max_shift']

		for node1, node2 in graph.edges:
			sorting1_name, unit_id1 = node1
			sorting2_name, unit_id2 = node2
			unit_idx1 = self.sortings[sorting1_name].id_to_index(unit_id1)
			unit_idx2 = self.sortings[sorting2_name].id_to_index(unit_id2)
			new_unit_id1 = self.aggregated_wvf_extractor.renamed_unit_ids[sorting1_name][unit_id1]
			new_unit_id2 = self.aggregated_wvf_extractor.renamed_unit_ids[sorting2_name][unit_id2]

			template1 = self.aggregated_wvf_extractor.get_template(new_unit_id1)
			template2 = self.aggregated_wvf_extractor.get_template(new_unit_id2)
			channel_indices = np.argsort(np.max(np.abs(template1) + np.abs(template2), axis=0))[:-n_channels-1:-1]

			shift = cross_shifts[sorting1_name][sorting2_name][unit_idx1, unit_idx2]
			template1 = template1[margin:-margin, channel_indices]
			template2 = template2[margin+shift:shift-margin, channel_indices]

			template_diff = np.sum(np.abs(template1 - template2)) / np.sum(np.abs(template1) + np.abs(template2))
			graph[node1][node2]['temp_diff'] = template_diff

	def clean_edges(self, graph: nx.Graph, cross_shifts: dict[str, dict[str, np.ndarray]], params: dict) -> None:
		"""
		Checks for edges that are bad (based on cross-contamination, template/correlogram difference)
		If an edge is bad, tries to find if one of the nodes is bad, and removes it
		If it cannot find a bad node, then it simply removes the edge.
		Nodes left alone (i.e. not connected) by this process are removed.

		@param graph: nx.Graph
			The graph containing all the units and connected by their similarity.
		@param cross_shifts: dict[str, dict[str, np.ndarray]]
			The cross-shifts between units.
		@param params: dict
			The parameters of the merge_sorting module.
		"""

		logs = open(f"{self.logs_folder}/clean_edges_logs.txt", 'w+')
		sd_threshold = 0.20  # TODO: Make these parameters.
		C_threshold = 0.08
		refractory_period = params['refractory_period']
		params = params['clean_edges']

		nodes_to_remove = []
		edges_to_remove = []

		for node1, node2, data in list(graph.edges(data=True)):
			sorting_name1, unit_id1 = node1
			sorting_name2, unit_id2 = node2

			unit_ind1 = self.sortings[sorting_name1].id_to_index(unit_id1)
			unit_ind2 = self.sortings[sorting_name2].id_to_index(unit_id2)

			C1 = graph.nodes[node1]['contamination']
			C2 = graph.nodes[node2]['contamination']

			sd1 = graph.nodes[node1]['sd_ratio']
			sd2 = graph.nodes[node2]['sd_ratio']

			spike_train1 = self.sortings[sorting_name1].get_unit_spike_train(unit_id1)
			spike_train2 = self.sortings[sorting_name2].get_unit_spike_train(unit_id2) + cross_shifts[sorting_name1][sorting_name2][unit_ind1, unit_ind2]
			if C2 < C1:
				spike_train1, spike_train2 = spike_train2, spike_train1

			cross_cont, p_value = utils.estimate_cross_contamination(spike_train1, spike_train2, refractory_period, limit=params['cross_cont_threshold'])

			problem_cases = [
				p_value < 5e-3 and data['temp_diff'] > params['template_diff_threshold'] and data['corr_diff'] > params['corr_diff_threshold'],
				p_value < 0.80 and data['similarity'] < 0.60 and data['temp_diff'] > 0.05 and data['corr_diff'] > 0.05
			]

			if not np.any(problem_cases):
				continue

			# From this point on, the edge is treated as problematic.
			logs.write(f"\nEdge {node1} -- {node2} is problematic:\n")
			logs.write(f"\t- {node1} - {graph.nodes[node1]}\n")
			logs.write(f"\t- {node2} - {graph.nodes[node2]}\n")
			logs.write(f"\t- cross-cont = {cross_cont:.2%} (p={p_value:.2e})\n")
			logs.write(f"\t- edge data: {data}\n")

			if sd1 - sd2 > (C2 - C1) * sd_threshold/C_threshold + sd_threshold and sd1 > 1.05:  # node 1 is problematic
				logs.write(f"\t=> Removing node {node1}\n")
				nodes_to_remove.append(node1)
			elif sd2 - sd1 > (C1 - C2) * sd_threshold/C_threshold + sd_threshold and sd2 > 1.05:  # node 2 is problematic
				logs.write(f"\t=> Removing node {node2}\n")
				nodes_to_remove.append(node2)
			else:  # Couldn't decide which one is problematic, Remove the connection between them.
				logs.write(f"\t=> Removing the connection\n")
				edges_to_remove.append((node1, node2))

		logs.write("\n\nComplete list:\n")
		for node in set(nodes_to_remove):
			logs.write(f"\t- Node {node} - {graph.nodes[node]}\n")
			graph.remove_node(node)
		for edge in edges_to_remove:
			if graph.has_edge(*edge):
				logs.write(f"\t- Edge {edge}\n")
				graph.remove_edge(*edge)

				for node in edge:  # If a node is left alone, remove it.
					if len(graph.edges(node)) == 0:
						logs.write(f"\t\t(also removed node {node} - {graph.nodes[node]} because left alone)\n")
						graph.remove_node(node)

		logs.close()

	@staticmethod
	def separate_communities(graph: nx.Graph) -> None:
		"""
		Looks for all subgraphs (connected component) and uses the Louvain algorithm to check if
		multiple communities are found. If so, the edges between communities are removed.
		Additionally, small communities are removed.
		Warning: it's recommended to run this function if there are at least 4 analyses.

		@param graph: nx.Graph
			The graph containing all the units and connected by their similarity.
		"""

		for nodes in list(nx.connected_components(graph)):
			subgraph = graph.subgraph(nodes)
			communities = list(nx.community.louvain_communities(subgraph, resolution=0.85))

			if len(communities) == 1:
				continue

			# Remove edges between communities
			for i in range(len(communities) - 1):
				for j in range(i+1, len(communities)):
					for node1 in communities[i]:
						for node2 in communities[j]:
							if graph.has_edge(node1, node2):
								graph.remove_edge(node1, node2)

			# Remove small communities
			for community in communities:
				if len(community) <= 2:
					for node in community:
						graph.remove_node(node)

	def merge_sortings(self, graph: nx.Graph, params: dict) -> si.NpzSortingExtractor:
		"""
		Merges the sortings based on a graph of similar units.

		@param graph: nx.Graph
			The graph containing all the units and connected by their similarity.
		@param params: dict
			The parameters of the merge_sorting module.
		@return merged_sorting: si.NpzSortingExtractor
			The merged sorting.
		"""

		k = 2.5
		t_c = int(round(params['refractory_period'][0] * 1e-3 * self.recording.sampling_frequency))
		new_spike_trains = {}
		logs = open(f"{self.logs_folder}/merge_sortings_logs.txt", 'w+')

		for nodes in nx.connected_components(graph):  # For each putative neuron.
			nodes = list(nodes)

			new_unit_id = len(new_spike_trains)
			best_score = -100000
			unit_label = ""
			logs.write(f"\nMaking unit {new_unit_id} from {nodes}\n")

			if len(nodes) == 1 and params['require_multiple_sortings_match']:  # Be more strict about nodes that end up being alone.
				node = nodes[0]
				if graph.nodes[node]['contamination'] > 0.10:
					logs.write(f"\t- Contamination too high (C = {graph.nodes[node]['contamination']:.1%})\n\t--> SKIPPING\n")
					continue
				if graph.nodes[node]['SNR'] < 3.0:
					logs.write(f"\t- SNR too low (SNR = {graph.nodes[node]['SNR']:.2f})\n\t--> SKIPPING\n")
					continue
				if abs(graph.nodes[node]['sd_ratio'] - 1) > 0.2:
					logs.write(f"\t- SD ratio too different from 1.0 (SD ratio = {graph.nodes[node]['sd_ratio']:.2f})\n\t--> SKIPPING\n")
					continue

			for n_units in range(1, params['max_units_merge']+1):
				for sub_nodes in itertools.combinations(nodes, n_units):
					spike_trains = [self.sortings[name].get_unit_spike_train(unit_id) for name, unit_id in sub_nodes]
					spike_train = np.sort(list(itertools.chain(*spike_trains))).astype(np.int64)
					indices_of_duplicates = find_duplicated_spikes(spike_train, t_c, method="keep_first_iterative")
					spike_train = np.delete(spike_train, indices_of_duplicates)

					f = len(spike_train) * self.sampling_f / self.recording.get_num_frames()
					C = utils.estimate_contamination(spike_train, params['refractory_period'])
					score = f * (1 - (k+1)*C)
					logs.write(f"\t- Score = {score:.2f}\t[{sub_nodes}]\n")

					if score > best_score:
						best_score = score
						new_spike_trains[new_unit_id] = spike_train
						unit_label = sub_nodes

			logs.write(f"\t--> Unit(s) chosen: {unit_label}\n")

		logs.close()

		sorting = si.NumpySorting.from_unit_dict(new_spike_trains, self.sampling_f)
		filename = f"{self.logs_folder}/merged_sorting.npz"
		si.NpzSortingExtractor.write_sorting(sorting, filename)

		merged_sorting = si.NpzSortingExtractor(filename)
		merged_sorting.annotate(name="merged_sorting")

		return merged_sorting



import itertools
import os
from typing import Any
import copy
import networkx as nx
from pathlib import Path
import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.postprocessing as spost


def get_path_to_plotlyJS(path: str | Path) -> Path:
	"""
	Returns the path to the plotly.js file relative to the given path.

	@param path: str | Path
		The path to where you are.
	@return plotlyJS_path: str
		The path to plotly.js relative to the given path.
	"""
	if isinstance(path, str):
		path = Path(path)

	return Path(os.path.relpath(utils.Utils.plotly_js_file, start=path))


def export_figure(fig: go.Figure, filepath: str) -> None:
	"""
	Exports a figure to html with the correct path to plotly.js.

	@param fig: Figure
		The figure to export.
	@param filepath: str
		The path where to save the plot.
	"""

	plotly_js = get_path_to_plotlyJS(Path(filepath).parent)
	fig.write_html(filepath, include_plotlyjs=plotly_js)


def plot_sliders(fig: go.Figure, traces_per_plot: int, labels: npt.ArrayLike, filepath: str, args: list[dict] | None = None, plots_per_file: int = 30) -> None:
	"""
	Takes a figure with multiple traces and plots, and decomposes it with a slider for each plot.
	If there are too many plots, will split the plots into multiple files.
	TODO: Add parameters for width, height, font_size etc.

	@param fig: go.Figure
		The plotly figure object.
	@param traces_per_plot: int
		Number of traces inside each plot.
	@param labels: ArrayLike
		The labels for each plot.
	@param filepath: str
		The path to the file to save the plot.
		Needs to be without any extension!
	@param args: list[dict] | None
		The arguments of what needs to change when the slider is moved.
	@param plots_per_file: int
		The maximum number of plots per file.
	"""
	if args is None:
		args = [{}] * len(labels)
	Path(filepath).parent.mkdir(parents=True, exist_ok=True)

	N = len(labels)
	n_files = 1 + (N - 1) // plots_per_file

	for n in range(n_files):
		start = (n * N) // n_files
		end = ((n+1) * N) // n_files

		sub_fig = copy.deepcopy(fig)
		sub_fig.data = sub_fig.data[traces_per_plot*start : traces_per_plot * end]
		for i in range(len(sub_fig.data)):
			sub_fig.data[i].visible = i < traces_per_plot

		steps = []
		for i in range(end - start):
			steps.append({
				'label': f"{labels[start+i]}",
				'method': "update",
				'args': [
					{'visible': [j // traces_per_plot == i for j in range(traces_per_plot * (end-start))]},
					args[start+i]
				]
			})

		sub_fig.layout['sliders'] = [{
			'active': 0,
			'currentvalue': {'prefix': 'Unit: '},
			'pad': {'t': 50},
			'steps': steps
		}]

		sub_fig.update_layout(**args[start])

		filename = f"{filepath}_{n+1}.html" if n_files > 1 else f"{filepath}.html"
		export_figure(sub_fig, filename)


def plot_units(wvf_extractor: si.WaveformExtractor, filepath: str, n_channels: int = 4, max_time_ms: float = 35., bin_size_ms: float = 0.25,
			   firing_rate_std: float = 8., annotations_fix: list[dict] | None = None, annotations_change: list[dict] | None = None) -> None:
	"""
	Plots all the units in a given sorting.

	@param wvf_extractor: si.WaveformExtractor
		The waveform extractor object containing the units to plot.
	@param filepath: str
		The path to the file to save the plot.
		Needs to be without any extension!
	@param n_channels: int
		The number of channels to plot for the template.
	@param max_time_ms: float
		The maximum time for ISI and auto-correlogram plots (in ms).
	@param bin_size_ms: float
		The bin size for ISI and auto-correlogram plots (in ms).
	@param firing_rate_std: float
		The standard deviation (in s) for the gaussian smoothing of the firing rate.
	@param annotations_fix: list[dict] | None
		The annotations that are fixed for all plots.
	@param annotations_change: list[dict] | None
		The annotations that change for each unit.
		Must be in the order [annot1_unit1, annot2_unit1, ... annot1_unit_2, annot2_unit2, ...]
	"""
	n_units = len(wvf_extractor.unit_ids)
	if n_units == 0:
		return

	if annotations_fix is None:
		annotations_fix = []
	if annotations_change is None:
		annotations_change = []
	assert len(annotations_change) % n_units == 0, "The number of annotations_change must be a multiple of the number of units!"
	n_annotations_per_plot = len(annotations_change) // n_units

	annotations_gt = create_gt_annotations(wvf_extractor.sorting)

	sf = wvf_extractor.sampling_frequency
	max_time = int(round(max_time_ms * 1e-3 * sf))
	bin_size = int(round(bin_size_ms * 1e-3 * sf))
	xaxis = (np.arange(wvf_extractor.nsamples) - wvf_extractor.nbefore) / sf * 1e3
	wvfs_unit = "µV" if wvf_extractor.return_scaled else "A.U."

	if n_channels > wvf_extractor.recording.get_num_channels():
		n_channels = wvf_extractor.recording.get_num_channels()

	spike_amplitudes = spost.compute_spike_amplitudes(wvf_extractor, load_if_exists=True, return_scaled=wvf_extractor.return_scaled, outputs="by_unit")[0]

	fig = go.Figure().set_subplots(rows=2+(n_channels-1)//4, cols=4)
	args = []

	for i, unit_id in enumerate(wvf_extractor.unit_ids):
		annotations_slice = slice(i*n_annotations_per_plot, (i+1)*n_annotations_per_plot)
		if i == 0:
			for annotation in annotations_fix:
				fig.add_annotation(**annotation)
			for annotation in annotations_change[annotations_slice]:
				fig.add_annotation(**annotation)
			if len(annotations_gt) > 0:
				fig.add_annotation(**annotations_gt[i])

		spike_train = wvf_extractor.sorting.get_unit_spike_train(unit_id)
		args.append({
			'title.text': f"Unit {unit_id}",
			'annotations': [*annotations_fix, *annotations_change[annotations_slice], *annotations_gt[i:i+1]]
		})

		bins = np.arange(bin_size/2, max_time, bin_size)
		ISI = np.histogram(np.diff(spike_train), bins=np.arange(0, max_time+bin_size, bin_size))[0]
		fig.add_trace(go.Bar(
			x=bins * 1e3 / sf,
			y=ISI,
			width=bin_size_ms,
			name="ISI",
			marker_color="CornflowerBlue"
		), row=1, col=1)

		auto_corr = spost.compute_autocorrelogram_from_spiketrain(spike_train, max_time, bin_size)
		bins, _, _ = spost.correlograms._make_bins(wvf_extractor.sorting, 2*max_time_ms, bin_size_ms)
		bin = (bins[1] - bins[0]) / 2
		fig.add_trace(go.Bar(
			x=bins[:-1] + bin,
			y=auto_corr,
			width=bin_size_ms,
			name="Auto-correlogram",
			marker_color="CornflowerBlue"
		), row=1, col=2)

		taxis = np.arange(0, wvf_extractor.recording.get_num_samples() + sf/2, sf / 2)
		firing_rate = utils.gaussian_histogram(spike_train, taxis, sigma=firing_rate_std * sf, margin_reflect=True) * sf
		fig.add_trace(go.Scatter(
			x=taxis / sf,
			y=firing_rate,
			mode="lines",
			name="Firing rate (Hz)",
			marker_color="CornflowerBlue"
		), row=1, col=3)

		fig.add_trace(go.Scattergl(
			x=wvf_extractor.sorting.get_unit_spike_train(unit_id) / sf,
			y=spike_amplitudes[unit_id],
			mode="markers",
			name="Spike amplitudes",
			marker_color="cornflowerblue"
		), row=1, col=4)

		template = wvf_extractor.get_template(unit_id, mode="average")
		best_channels = np.argsort(np.max(np.abs(template), axis=0))[::-1]
		for i in range(n_channels):
			channel = best_channels[i]
			fig.add_trace(go.Scatter(
				x=xaxis,
				y=template[:, channel],
				mode="lines",
				name=f"Template channel {wvf_extractor.channel_ids[channel]}",
				marker_color="CornflowerBlue"
			), row=2 + i//4, col=1 + i%4)

	fig.update_xaxes(title_text="Time (ms)", row=1, col=1)
	fig.update_xaxes(title_text="Time (ms)", row=1, col=2)
	fig.update_xaxes(title_text="Time (s)", row=1, col=3)
	fig.update_xaxes(title_text="Time (s)", row=1, col=4)
	fig.update_yaxes(title_text="ISI", rangemode="tozero", row=1, col=1)
	fig.update_yaxes(title_text="Auto-correlogram", rangemode="tozero", row=1, col=2)
	fig.update_yaxes(title_text="Firing rate (Hz)", rangemode="tozero", row=1, col=3)
	fig.update_yaxes(title_text=f"Spike amplitudes ({wvfs_unit})", rangemode="tozero", row=1, col=4)
	for i in range(n_channels):
		fig.update_xaxes(title_text="Time (ms)", matches='x5', row=2 + i//4, col=1 + i%4)
		fig.update_yaxes(title_text=f"Voltage ({wvfs_unit})", rangemode="tozero", matches='y5', row=2 + i//4, col=1 + i%4)

	plot_sliders(fig, 4 + n_channels, labels=wvf_extractor.unit_ids, filepath=filepath, args=args)


def create_gt_annotations(sorting: si.BaseSorting) -> list[dict[str, Any]]:
	"""
	Creates the Ground Truth annotations for the plot.
	Returns an empty array if property 'gt_label' is not found in the sorting object.

	@param sorting: BaseSorting
		The sorting for which to create the annotations.
	@return gt_annotations: list[dict[str, Any]]
		The Ground Truth annotations for the plot.
	"""

	annotations_gt = []

	if 'gt_label' in sorting.get_property_keys():
		for unit_id in sorting.unit_ids:
			gt_label = sorting.get_unit_property(unit_id, 'gt_label')
			annotations_gt.append({
				'x': 1.0,
				'y': 1.05,
				'xref': 'paper',
				'yref': 'paper',
				'xanchor': 'right',
				'yanchor': 'top',
				'text': f"GT: {gt_label}",
				'font': {
					'color': "rgba(73, 42, 189, 1.0)"
				},
				'showarrow': False
			})

	return annotations_gt


def create_graph_plot(graph: nx.Graph) -> go.Figure:  # pragma: no cover (test function)
	nodes_pos = nx.spring_layout(graph)

	fig = go.Figure()

	fig.add_trace(go.Scatter(
		x=list(itertools.chain(*[[nodes_pos[node1][0], nodes_pos[node2][0], None] for node1, node2 in graph.edges])),
		y=list(itertools.chain(*[[nodes_pos[node1][1], nodes_pos[node2][1], None] for node1, node2 in graph.edges])),
		mode="lines",
		line=dict(width=1, color="rgba(0, 0, 0, 0.8)"),  # TODO: color problematic edges in red.
		showlegend=False
	))

	fig.add_trace(go.Scatter(
		x=np.array(list(nodes_pos.values()))[:, 0],
		y=np.array(list(nodes_pos.values()))[:, 1],
		mode="markers",
		marker_color=["Crimson" if 'problem' in data and data['problem'] else "CornflowerBlue" for _, data in graph.nodes(data=True)],
		text=[(f"{name}<br />" + "<br />".join(f"{key}: {value}" for key, value in data.items())) for name, data in graph.nodes(data=True)]
	))

	return fig
